"""
A backend that talks using MQTT
"""

from __future__ import annotations

import anyio
import logging
import time
from contextlib import asynccontextmanager

from mqttproto.async_client import AsyncMQTTClient, PropertyType, RetainHandling, Will

from moat.util import NotGiven, attrdict, get_codec
from moat.lib.codec.noop import Codec as NoopCodec
from moat.link.meta import MsgMeta
from moat.util.path import PS, P, Path

from . import Backend as _Backend
from . import Message, RawMessage

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from moat.lib.codec import Codec

    from collections.abc import Any, AsyncIterator, Awaitable, Literal


class MqttMessage:
    """
    Encapsulates (our view of) a message from MQTT.
    """

    def __init__(self, topic, payload, orig, **kw):
        self.topic = topic
        self.payload = payload
        self.orig = orig
        self.meta = attrdict(kw)


class Backend(_Backend):
    """
    The MQTT backend driver.
    """

    client: AsyncMQTTClient

    def __init__(
        self,
        cfg,
        will: attrdict | None = None,
        name: str | None = None,
        id: str | None = None,
        meta: bool = True,
    ):
        """
        Connect to MQTT.

        @cfg: at moat.link.backend
        @meta: if set (the default), always attach metadata when sending.
        @will: topic+data+retain+qos+codec for on-death message

        """
        super().__init__(cfg, name=name, id=id)
        self.cfg = cfg
        self.meta = meta
        self.logger = logging.getLogger(__name__ + "." + (self.name or "‹…›"))

        kw = cfg.copy()
        sname = kw.pop("driver", "mqtt")
        self.trace = kw.pop("trace", False)
        try:
            codec = kw.pop("codec")  # intentionally no default
        except KeyError:
            raise RuntimeError(f"The {sname} backend requires a codec.") from None

        kw["client_id"] = self.name

        a = (kw.pop("host"),) if "host" in kw else ()
        self.codec = get_codec(codec)
        self.mcodec = get_codec("std-cbor")

        will = will or kw.pop("will", None)
        if will is not None:
            data = will.pop("data", NotGiven)
            cdc = will.pop("codec", NotGiven)
            cdc = self.codec if cdc is NotGiven else get_codec(cdc)

            data = b"" if data is NotGiven else cdc.encode(data)
            kw["will"] = Will(
                topic=will["topic"].slashed2,
                payload=data,
                qos=will.get("qos", 1),
                retain=will.get("retain", data == b""),
            )
        self.a, self.kw = a, kw

    @asynccontextmanager
    async def connect(self):
        "connect to the server"
        async with AsyncMQTTClient(*self.a, **self.kw) as self.client:
            try:
                yield self
            finally:
                self.client = None  # noqa:PLW2901

    @asynccontextmanager
    async def monitor(
        self,
        topic,
        *,
        codec: str | Codec | None | Literal[NotGiven] = NotGiven,
        raw: bool | None = False,
        subtree: bool = False,
        mine: bool = True,
        retained: bool = True,
        **kw,
    ) -> AsyncIterator[AsyncIterator[Message]]:
        """
        Watch a topic.

        @codec: use this codec.
        @raw: don't interpret anything.
        @subtree: also monitor subtopics.
        @mine: send my own messages back to me.
        """

        if not topic.is_empty:
            tops = topic.slashed
            if subtree:
                tops += "/#"
        elif subtree:
            tops = "#"
        else:
            raise ValueError("empty path")
        self.logger.debug("Monitor %s start", tops)
        codec = self.codec if codec is NotGiven else get_codec(codec)
        kw["no_local"] = not mine
        kw["retain_handling"] = (
            RetainHandling.SEND_RETAINED if retained else RetainHandling.NO_RETAINED
        )
        try:
            async with self.client.subscribe(tops, **kw) as sub:
                yield _SubGet(self, sub, codec, raw)
        except (anyio.get_cancelled_exc_class(), KeyboardInterrupt):
            raise
        except BaseException as exc:
            self.logger.exception("Monitor %s end", topic, exc_info=exc)
            raise
        else:
            self.logger.debug("Monitor %s end", topic)

    @overload
    def send(
        self,
        topic: Path,
        data: bytes | bytearray | memoryview,
        codec: Literal[None],
        meta: MsgMeta | bool | None = None,
        retain: bool | None = None,
        **kw,
    ) -> Awaitable:  # pylint: disable=invalid-overridden-method
        ...

    @overload
    def send(
        self,
        topic: Path,
        data: Any,
        codec: Codec | str | Literal[NotGiven] = NotGiven,
        meta: MsgMeta | bool | None = None,
        retain: bool | None = None,
        **kw,
    ) -> Awaitable:  # pylint: disable=invalid-overridden-method
        ...

    def send(
        self,
        topic: Path,
        data: Any,
        codec: Codec | str | None | Literal[NotGiven] = NotGiven,
        meta: MsgMeta | bool | None = None,
        retain: bool | None = None,
        **kw,
    ) -> Awaitable:  # pylint: disable=invalid-overridden-method
        """
        Send this payload to this topic.

        The keyword arguments @timestamp and @origin will be used for the
        MoaT user property.
        """
        prop = {}
        if meta is None:
            meta = self.meta
        if meta is False:
            if retain is None:
                retain = False
        else:
            if meta is True:
                meta = MsgMeta(origin=self.name)
            prop["MoaT"] = meta.encode()
        if retain is None:
            raise ValueError("Need to set whether to retain or not")

        if isinstance(data, str):
            msg = data  # utf-8 is pass-thru in MQTT5
        elif data is NotGiven:
            # delete
            msg = b""
        else:
            if codec is NotGiven:
                codec = self.codec
            elif codec is None:
                codec = NoopCodec()
            elif isinstance(codec, str):
                codec = get_codec(codec)
            # else codec is a Codec and used as-is
            msg = codec.encode(data)

        if self.trace:
            self.logger.info("S:%s %r", topic, data)
        return self.client.publish(
            topic.slashed2,
            payload=msg,
            user_properties=prop,
            retain=retain,
            **kw,
        )


class _SubGet:
    def __init__(self, back, sub, codec, raw):
        self.back = back
        self.sub = sub
        self.codec = codec
        self.raw = raw

    def __aiter__(self):
        return self

    async def __anext__(self):
        back = self.back
        err = None
        while True:
            msg = await anext(self.sub)
            try:
                top = PS(msg.topic)
            except Exception as exc:
                await back.send(
                    P(":R.error.link.mqtt.topic"),
                    dict(val=msg.topic, msg=repr(exc)),
                    retain=False,
                )
                # workaround for undecodeability
                top = Path.build(msg.topic.split("/"))

            prop = msg.user_properties.get("MoaT")
            if not self.raw:
                oprop = prop  # remember for error
                try:
                    if prop:
                        prop = MsgMeta.decode(back.name, prop)
                    else:
                        prop = MsgMeta(name=back.name)

                    assert prop.origin
                    if not prop.timestamp:
                        prop.timestamp = time.time()

                except Exception as exc:
                    back.logger.debug("Property Error", exc_info=exc)
                    await back.send(
                        P(":R.error.link.mqtt.meta") + top,
                        dict(
                            topic=top,
                            val=oprop,
                            msg=repr(exc),
                            retain=False,
                        ),
                    )
                    err = exc
                else:
                    try:
                        p_i = msg.properties.get(PropertyType.PAYLOAD_FORMAT_INDICATOR, 0)
                        if not p_i:
                            data = self.codec.decode(msg.payload)
                        elif p_i == 1:
                            data = msg.payload  # UTF-8
                        else:
                            raise ValueError("Unknown payload format {p_i}")
                    except Exception as exc:
                        back.logger.warning(
                            "Decoding Error %s %s: %r %r",
                            top,
                            self.codec.__class__.__module__,
                            msg.payload,
                            exc,
                            exc_info=exc,
                        )
                        await back.send(
                            P(":R.error.link.mqtt.codec") + top,
                            dict(
                                codec=type(self.codec).__name__,
                                topic=top,
                                val=msg.payload,
                                msg=repr(exc),
                            ),
                            retain=False,
                        )
                        err = exc
                    else:
                        # everything OK
                        if back.trace:
                            back.logger.debug("R:%s %r", top, data)
                        return Message(
                            top, data, meta=prop, prop=msg.user_properties, retain=msg.retain
                        )
                    continue
            if self.raw is False:
                # don't forward undecodeable messages
                continue
            if back.trace:
                back.logger.info("R:%s R|%r", top, msg.payload)
            return RawMessage(
                top, msg.payload, meta=prop, prop=msg.user_properties, exc=err, retain=msg.retain
            )



"""
This module describes the MoaT data link.
"""

from __future__ import annotations

import anyio

# pylint: disable=missing-module-docstring
import logging
import sys

import asyncclick as click

from moat.util import NotGiven, P, Path, load_subgroup, yprint
from moat.util.path import set_root

from .backend import get_backend

try:
    from mqttproto import MQTTPublishPacket
except ImportError:
    MQTTPublishPacket = NotGiven


logger = logging.getLogger(__name__)

usage1 = """
"moat link" requires configuration. It should look like this:

    link:
        backend:
            driver: mqtt
            host: localhost
            port: 1883
            user: foo
            pass: bar
            codec: std-cbor
        root: !P moat.something-unique

The MQTT hierarchy below the "root" topic, with slashes instead of dots,
*must* be empty when you start a new installation.

Please add this stanza to the file "/etc/moat/moat.cfg" (or "/etc/moat.cfg",
or "~/.config/moat.cfg") and try again.
"""

usage2 = """
"moat link" requires a root topic.

This entry should be tagged with '!P' and be a dot-separated path.
For clarity it should start with 'moat.', though that's just a
recommendation.

Your MQTT setup may NOT use any non-MoaT topics under that name
(replacing the dots with slashes).

Config example:

    link:
        backend:
            …
        prefix: !P moat.com.your-domain

"""

usage9 = """

"moat link" requires at least one history server for stable operation.

Please run "sudo systemctl start moat-link-server", or
start "moat link server" in a separate terminal, and try again.
"""


@load_subgroup(sub_pre="moat.link.cmd", sub_post="cli", ext_pre="moat.link", ext_post="_main.cli")
@click.pass_context
async def cli(ctx):
    """
    MoaT's data link

    Commands for running and managing MoaT's data link layer.
    """
    obj = ctx.obj
    cfg = obj.cfg

    if "link" not in cfg or "backend" not in cfg["link"]:
        sys.stderr.write(usage1)
        raise click.UsageError("not configured")
    if not isinstance(cfg.link.root, Path) or cfg.link.root == P("XXX.NotConfigured.YZ"):
        sys.stderr.write(usage2)
        raise click.UsageError("badly configured")


@cli.command("test")
@click.pass_obj
async def test(obj):
    "Test"

    lock = anyio.Lock()
    cfg = obj.cfg.link
    set_root(cfg)

    async def check_root():
        try:
            with anyio.fail_after(1):
                async with back.monitor(cfg.root) as mon:
                    async for msg in mon:
                        async with lock:
                            print("# Retained root dataset:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"## No retained root dataset on {cfg.root}.")

    async def check_server():
        try:
            with anyio.fail_after(1):
                async with back.monitor(P(":R.run.service.main.conn")) as mon:
                    async for msg in mon:
                        async with lock:
                            print("# Server link:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"### No server link on {cfg.root}!")

    async with get_backend(cfg) as back, back.connect(), anyio.create_task_group() as tg:
        tg.start_soon(check_root)
        tg.start_soon(check_server)



import httpx
from asyncakumuli import get_data
import asyncclick as click
import datetime
import time
from moat.util import P, yload, CFG, ensure_cfg, merge
import sys
from moat.link.client import Link
import anyio

series = "price"
tags = dict(type="power", source="awattar", redo="A")


def sdate(d):
    d = datetime.datetime.fromtimestamp(d)
    return d.strftime("%Y-%m-%d %H:%M")


@click.command()
@click.option("--path", "-p", type=P, help="Value to set")
@click.option("--verbose", "-v", is_flag=True, help="Report values as set")
@click.option(
    "--config", "-c", type=click.File("r"), help="MoaT config file", default="/etc/moat/moat.cfg"
)
@click.option("--offset", "-o", type=float, help="offset from start (seconds)", default=0)
async def main(path, verbose, config, offset):
    """
    Feed a stored Akumuli time series back to MoaT-KV.
    """
    if path is None and not verbose:
        raise click.UsageError("No path and no verbosity is a no-op")
    with config as f:
        cfg = yload(f, attr=True)
    ensure_cfg("moat.link")
    merge(CFG, cfg)
    tt = time.time()
    val = None
    async with Link(CFG["link"]) as cli, httpx.AsyncClient() as s:
        while True:
            seen = False
            t = datetime.datetime.now()
            async for v in get_data(
                s,
                series,
                tags,
                url="http://a.rock.s:50081/api/query",
                t_start=t - datetime.timedelta(0, 10000 - offset),
                t_end=t + datetime.timedelta(2, offset),
            ):
                if v.time + offset < tt:
                    print("Skip:", v.time)
                    val = v.value
                    continue
                if val is not None:
                    if verbose:
                        print(val)
                        sys.stdout.flush()
                    if path is not None:
                        await cli.d_set(path, val, retain=True)
                elif verbose:
                    print("Series starts in the future", file=sys.stderr)
                print("Sleep:", v.time + offset - tt, "to", sdate(v.time + offset))
                await anyio.sleep(v.time + offset - tt)
                tt = time.time()
                val = v.value
                seen = True
            if not seen:
                print("No data, sleeping", file=sys.stderr)
                await anyio.sleep(3000)


main(_anyio_backend="trio")



from __future__ import annotations  # noqa: D100

import anyio
import logging
import time
from contextlib import asynccontextmanager, suppress
from enum import Enum, auto
from functools import partial

from attrs import define, field
from transitions_aio.extensions.factory import MachineFactory

from moat.util import CtxObj, NotGiven, P, attrdict, srepr
from moat.lib.priomap import PrioMap
from moat.util.broadcast import Broadcaster

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from transitions_aio import EventData

    from moat.util import Path
    from moat.link.client import Link

    from collections.abc import Awaitable

__all__ = ["HostList", "ServiceMon"]

logger = logging.getLogger(__name__)


#
# Monitoring
#
class HostState(Enum):
    # constants
    INIT = "in"
    NEW = "nw"
    DOWN = "dn"  # id state=down seen
    UP = "up"
    TIMEOUT = "tm"  # no ping for some time, error generated
    STALE = "sl"  # superseded
    DROP = "xx"  # no ping for a long time, entry dropped

    # We want "id" and "ping" messages.
    # "host" is an additional "name this host" signal.
    ONLY_I = "oi"  # only host/id seen
    ONLY_P = "op"  # only ping seen


class HostEvent(Enum):
    INIT = auto()
    MSG_HOST = auto()
    MSG_PING = auto()
    MSG_ID = auto()

    MSG_DOWN = auto()
    DEL_HOST = auto()  # Service gets dropped, or replaced
    DEL_ID = auto()

    TIMEOUT = auto()


_S = HostState
_E = HostEvent


# @add_state_features(MonTimeout)
class HostMachine(MachineFactory.get_predefined(graph=True, asyncio=True)):
    def __init__(self, host):
        states = [
            dict(name=_S.INIT),
            dict(name=_S.NEW),
            dict(name=_S.DOWN),
            dict(name=_S.UP),
            dict(name=_S.TIMEOUT),
            dict(name=_S.STALE),
            dict(name=_S.DROP, final=True),
            dict(name=_S.ONLY_I),
            dict(name=_S.ONLY_P),
        ]
        transitions = [
            dict(trigger=_E.INIT, source=_S.INIT, dest=_S.NEW),
            dict(trigger=_E.MSG_HOST, source=_S.NEW, dest=_S.ONLY_I),
            dict(trigger=_E.MSG_ID, source=_S.NEW, dest=_S.ONLY_I),
            dict(trigger=_E.MSG_PING, source=_S.NEW, dest=_S.ONLY_P),
            dict(trigger=_E.MSG_DOWN, source=_S.NEW, dest=_S.DROP),
            dict(trigger=_E.DEL_HOST, source=_S.NEW, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.NEW, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.NEW, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.ONLY_I, dest=_S.ONLY_I),
            dict(trigger=_E.MSG_ID, source=_S.ONLY_I, dest=_S.ONLY_I),
            dict(trigger=_E.MSG_PING, source=_S.ONLY_I, dest=_S.UP),
            dict(trigger=_E.MSG_DOWN, source=_S.ONLY_I, dest=_S.DROP),
            dict(trigger=_E.DEL_HOST, source=_S.ONLY_I, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.ONLY_I, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.ONLY_I, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.ONLY_P, dest=_S.ONLY_P),
            dict(trigger=_E.MSG_ID, source=_S.ONLY_P, dest=_S.UP),
            dict(trigger=_E.MSG_PING, source=_S.ONLY_P, dest=None),
            dict(trigger=_E.MSG_DOWN, source=_S.ONLY_P, dest=_S.DROP),
            dict(trigger=_E.DEL_HOST, source=_S.ONLY_P, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.ONLY_P, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.ONLY_P, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.DOWN, dest=_S.DOWN),
            dict(trigger=_E.MSG_ID, source=_S.DOWN, dest=_S.DOWN),
            dict(trigger=_E.MSG_PING, source=_S.DOWN, dest=_S.DOWN),
            dict(trigger=_E.MSG_DOWN, source=_S.DOWN, dest=_S.DROP),
            dict(trigger=_E.DEL_HOST, source=_S.DOWN, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.DOWN, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.DOWN, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.UP, dest=_S.UP),
            dict(trigger=_E.MSG_ID, source=_S.UP, dest=_S.UP),
            dict(trigger=_E.MSG_PING, source=_S.UP, dest=None),
            dict(trigger=_E.MSG_DOWN, source=_S.UP, dest=_S.DOWN),
            dict(trigger=_E.DEL_HOST, source=_S.UP, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.UP, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.UP, dest=_S.STALE),
            dict(trigger=_E.MSG_HOST, source=_S.TIMEOUT, dest=_S.TIMEOUT, after=host.drop_host),
            dict(trigger=_E.MSG_ID, source=_S.TIMEOUT, dest=_S.TIMEOUT, after=host.drop_id),
            dict(trigger=_E.MSG_PING, source=_S.TIMEOUT, dest=_S.TIMEOUT, after=host.drop_id),
            dict(trigger=_E.MSG_DOWN, source=_S.TIMEOUT, dest=_S.DROP),
            dict(trigger=_E.DEL_HOST, source=_S.TIMEOUT, dest=_S.DROP),
            dict(trigger=_E.DEL_ID, source=_S.TIMEOUT, dest=_S.DROP),
            dict(trigger=_E.TIMEOUT, source=_S.TIMEOUT, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.STALE, dest=None),
            dict(trigger=_E.MSG_ID, source=_S.STALE, dest=None),
            dict(trigger=_E.MSG_PING, source=_S.STALE, dest=_S.UP),
            dict(trigger=_E.MSG_DOWN, source=_S.STALE, dest=None),
            dict(trigger=_E.DEL_HOST, source=_S.STALE, dest=None),
            dict(trigger=_E.DEL_ID, source=_S.STALE, dest=None),
            dict(trigger=_E.TIMEOUT, source=_S.STALE, dest=_S.DROP),
            dict(trigger=_E.MSG_HOST, source=_S.DROP, dest=None, after=host.re_init),
            dict(trigger=_E.MSG_ID, source=_S.DROP, dest=None, after=host.re_init),
            dict(trigger=_E.MSG_PING, source=_S.DROP, dest=None, after=host.re_init),
            dict(trigger=_E.MSG_DOWN, source=_S.DROP, dest=None, after=host.re_init),
            dict(trigger=_E.DEL_HOST, source=_S.DROP, dest=None, after=host.re_init),
            dict(trigger=_E.DEL_ID, source=_S.DROP, dest=None, after=host.drop_id),
            dict(trigger=_E.TIMEOUT, source=_S.DROP, dest=None, after=host.drop_both),
        ]
        for s in states:
            sn = s["name"].name.lower()
            if (cb := getattr(host, f"on_enter_{sn}", None)) is not None:
                s["on_enter"] = [cb]
            if (cb := getattr(host, f"on_exit_{sn}", None)) is not None:
                s["on_exit"] = [cb]

        super().__init__(
            # model=host,
            states=states,
            initial=HostState.INIT,
            transitions=transitions,
            auto_transitions=False,
            ignore_invalid_triggers=False,
            send_event=True,
            queued=True,
            after_state_change=[self._set_timeout, host.updated],
            on_final=host.drop_both,
        )
        self.host = host

    def _set_timeout(self, ed: EventData):
        host = ed.model.host
        host.mon.set_timeout(host)


@define(eq=False)
class Service:
    """Contains data for one particular service."""

    mon: HostList = field()
    id: str = field()

    # These get filled via path
    data: dict[str, dict] = field(factory=attrdict, init=False)
    machine: HostMachine = field(init=False)

    _last = field(default=0, init=False)

    def __attrs_post_init__(self):
        self.machine = HostMachine(self)
        self.data.h = dict()

    @property
    def state(self):
        return self.machine.state

    @property
    def timeout(self):
        "returns the timeout under `cfg.link.timout.ping`"
        if self.state is HostState.DROP:
            return 0  # error when looked up

        if self.state in (HostState.ONLY_I, HostState.TIMEOUT):
            val = "stale"
        elif self.state in (HostState.DOWN, HostState.STALE):
            val = "delete"
        else:
            val = "timeout"
        res = self.mon.cfg.timeout.ping[val]
        if self.last:
            age = time.time() - self.last
            res = min(res, max(self.mon.cfg.timeout.ping.min, age))
        return res

    def updated(self, evt):
        "queue update to Broadcaster"
        self.mon.updated(self, evt)

    def drop_id(self) -> Awaitable[None]:
        return self.mon.drop_id(self)

    def drop_host(self) -> Awaitable[None]:
        return self.mon.drop_host(self)

    async def drop_both(self, evt=None) -> None:
        evt  # noqa:B018
        await self.mon.drop_host(self)
        await self.mon.drop_id(self)

    @property
    def last(self):
        return self._last

    @last.setter
    def last(self, tm):
        if self._last < tm:
            self._last = tm

    def trigger(self, *args, **kwargs) -> Awaitable:
        return self.machine.trigger(*args, **kwargs)

    async def on_enter_drop(self, ev):  # noqa: ARG002
        await self.mon.drop_cb(self)

    async def re_init(self, ev):
        """
        We have an incoming packet while a DROP event is running.
        This should not happen, but whatever …
        """
        ho = ev.model.host
        mon = ho.mon
        # dprint("**RE*",ho.id)
        h = Service(mon, ho.id)
        h.data = ho.data
        # not copying the paths!
        mon.ids[h.id] = h
        await h.trigger(_E.INIT)


class HostList(CtxObj):
    """
    This class monitors the 'run.ping', 'run.id' and 'run.host' channels.
    """

    def __init__(
        self, cfg: dict, link: Link, debug: bool = False, broadcaster: Broadcaster = None
    ):
        self.link = link
        self.debug = debug
        self.cfg = cfg
        self._bc = broadcaster

        self._retime = anyio.Event()

        # Service Instance
        self.hsi: dict[Path, Service] = {}
        self.ids: dict[str, Service] = {}
        self.times: PrioMap[str, float] = PrioMap()

    @asynccontextmanager
    async def _ctx(self):
        if self._bc is None:
            self._bc = Broadcaster(10)

        async with self._bc, anyio.create_task_group() as self.tg:
            self.tg.start_soon(self._mon_host)
            self.tg.start_soon(self._mon_ping)
            self.tg.start_soon(self._mon_id)
            self.tg.start_soon(self._timer)
            yield self._bc
            self.tg.cancel_scope.cancel()

    def set_timeout(self, host: Service, to: float | None = None):
        "Update service entry in timeout queue"
        if to is None:
            to = host.timeout
        if not to:
            self.times.pop(host.id, None)
            return

        tm = time.monotonic() + to

        if len(self.times) and self.times.peek()[1] > tm:
            self._retime.set()
        # dprint("TIME ",host.id,tm-time.monotonic())
        self.times[host.id] = tm

    async def _timer(self):
        async for id, timeout in self.times:
            h = self.ids[id]

            t = time.monotonic()
            if t < timeout:
                self.times[id] = timeout
                # dprint("WAIT+",id,timeout-t)
                with anyio.move_on_after(timeout - t):
                    await self._retime.wait()
                    self._retime = anyio.Event()
                continue

            # dprint("OUT  ",h.id)
            self.tg.start_soon(h.trigger, _E.TIMEOUT)

    def updated(self, host: Service, evt: EventData):
        "queue update to Broadcaster"
        if evt.transition.dest is not None or evt.kwargs.get("changed", False):
            logger.debug("Service %s: %s %s", host.id, host.state.name, srepr(host.data))
            self._bc(host)

    async def _mon_host(self):
        """
        Monitor host messages.

        These update the .hsi lookup but don't actually affect state.
        """
        async with self.link.d_watch(
            P("run.host"), subtree=True, meta=True, state=NotGiven
        ) as mon:
            async for p, msg, meta in mon:
                try:
                    if msg is NotGiven:  # host deleted
                        with suppress(KeyError, AttributeError):
                            h = self.hsi.pop(p)
                            h.data.h.pop(p, None)
                            if not h.data.h:
                                self.tg.start_soon(h.trigger, _E.DEL_HOST)

                        continue

                    id = msg.pop("id")

                    # points to another service? drop it from the other entry
                    with suppress(KeyError, AttributeError):
                        h = self.hsi[p]
                        if h.id != id:
                            h.data.h.pop(p, None)
                            if not h.data.h:
                                self.tg.start_soon(h.trigger, _E.DEL_HOST)
                            # unconditionally updated below

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)
                    self.hsi[p] = h
                    h.last = meta.timestamp
                    changed = h.data.h.get(p, None) != msg
                    h.data.h[p] = msg
                    self.tg.start_soon(partial(h.trigger, _E.MSG_HOST, changed=changed))
                    # dprint("H    ",p,id,h)

                except Exception as exc:
                    logger.error("BadHost %r %r", p, msg, exc_info=exc)

    async def _mon_id(self):
        async with self.link.d_watch(P("run.id"), subtree=True, meta=True, state=NotGiven) as mon:
            async for p, msg, meta in mon:
                try:
                    if len(p) != 1:
                        logger.warning("BadID %r %r", p, msg)
                        continue

                    id = p[0]
                    if msg is NotGiven:  # ID deleted
                        with suppress(KeyError):
                            h = self.ids[id]
                            self.tg.start_soon(h.trigger, _E.DEL_ID)
                        continue

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)
                    h.last = meta.timestamp

                    changed = False
                    if h.data.get("i", None) != msg:
                        h.data.i = msg
                        changed = True
                    self.tg.start_soon(partial(h.trigger, _E.MSG_ID, changed=changed))
                    # dprint("P    ",p,msg)

                except Exception as exc:
                    logger.error("BadId %r %r", p, msg, exc_info=exc)

    async def _mon_ping(self):
        async with self.link.d_watch(P("run.ping.id"), subtree=True, state=NotGiven) as mon:
            async for p, msg in mon:
                try:
                    if len(p) != 1 or not isinstance(msg, dict):
                        logger.warning("BadPING %r %r", p, msg)
                        continue
                    id = p[0]

                    try:
                        h = self.ids[id]
                    except KeyError:
                        self.ids[id] = h = Service(mon=self, id=id)
                        await h.trigger(_E.INIT)

                    changed = False
                    if h.data.get("p", None) != msg:
                        h.data.p = msg
                        changed = True
                    self.tg.start_soon(
                        partial(
                            h.trigger, _E.MSG_PING if msg["up"] else _E.MSG_DOWN, changed=changed
                        )
                    )
                    # dprint("P    ",p,msg)

                except Exception as exc:
                    logger.error("BadPing %r %r", p, msg, exc_info=exc)

    async def drop_cb(self, host):
        """
        Called from `Service.on_enter_drop`
        """
        # dprint("  DCB",host.id)
        for p in host.data.h:
            if self.hsi.get(p, None) is host:
                del self.hsi[p]
        if self.ids.get(host.id, None) is host:
            del self.ids[host.id]
        self.set_timeout(host, False)

    async def drop_id(self, host):
        """
        Send a message to delete this ID entry.
        """
        pass

    async def drop_host(self, host):
        """
        Send a message to delete this host entry.
        """
        pass


class ServiceMon(HostList):
    """
    The Service Monitor runs once in a MoaT-Link network, as part of the main
    hosts's 'moat-link-host' service.

    Its main job is to remove stale retained entries under 'run.id' and 'run.host'.
    """

    async def drop_id(self, host):
        """
        Send a message to delete this ID entry.
        """
        await self.link.d_set(P("run.id") / host.id, retain=True)
        await super().drop_id(host)

    async def drop_host(self, host):
        """
        Send a message to delete this host entry.
        """
        for p in host.data.h.keys():
            await self.link.d_set(P("run.host") + p, retain=True)
        await super().drop_host(host)



"""
This module forwards notifications to NTFY.
"""

from __future__ import annotations

from contextlib import asynccontextmanager

import httpx

from moat.util import P, Path

from . import Notifier as BaseNotifier

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = ["Notifier"]

prio_map = {
    "debug": 1,
    "info": 2,
    "warning": 3,
    "error": 4,
    "fatal": 5,
}


class Notifier(BaseNotifier):  # noqa:D101
    @asynccontextmanager
    async def _ctx(self):
        async with (
            super()._ctx(),
            httpx.AsyncClient() as self.http,
        ):
            yield self

    async def send(
        self,
        topic: str | Path,
        title: str,
        msg: str,
        prio: str | None = None,
        tags: Sequence[str] = (),
        **kw,  # noqa: ARG002
    ):
        "Forward a message to NTFY"

        if "topic" in self.cfg:
            top = self.cfg.topic
        else:
            if isinstance(topic, str):
                topic = P(topic)
            try:
                top = topic[0]
            except IndexError:
                top = "TOP"
            else:
                topic = Path.build(topic[1:])

        url = self.cfg.url + "/" + top
        tags = list(tags)
        tags.append(str(topic))

        hdr = {}

        if title:
            hdr["title"] = title
        if prio is not None:
            try:
                prio = prio_map[prio]
                hdr["prio"] = str(prio)
                if prio > 3:
                    tags.append("warning")
            except KeyError:
                hdr["prio"] = "high"
                msg += f" (prio:{prio})"
        hdr["tags"] = ",".join(str(x) for x in tags)

        if "token" in self.cfg:
            hdr["authorization"] = f"Bearer {self.cfg['token']}"

        await self.http.post(url, data=msg, headers=hdr)



"""
MoaT gateway
"""

from __future__ import annotations

import anyio
from contextlib import AsyncExitStack

from moat.util import NotGiven, P, Path
from moat.link.meta import MsgMeta

from . import Gate as _Gate

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.link.node.codec import CodecNode

    from . import GateNode

    from typing import Any


class Gate(_Gate):  # noqa: D101
    codecs: CodecNode | None = None

    async def run_(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        "Main loop. Overridden to fetch the codecs"
        async with AsyncExitStack() as ex:
            if isinstance(self.cf.codec, Path):
                cdv = await ex.enter_async_context(
                    self.link.d_watch(
                        P("conv") + self.cf.codec, subtree=True, state=None, meta=False
                    )
                )
                self.codec_vecs = await cdv.get_node()

                self.codecs = await self.link.get_codec_tree()

            await super().run_(task_status=task_status)

    async def get_dst(self, *, task_status=anyio.TASK_STATUS_IGNORED):  # noqa: D102
        async with AsyncExitStack() as ex:
            if self.codecs is not None:
                codec = "noop"

                def conv(p, d):
                    # two step
                    # (a) look up the codec type in the vector
                    try:
                        vd = self.codec_vecs.search(p)
                        cd = self.codecs.get(Path.build(vd.data["codec"]))
                    except (KeyError, ValueError):
                        return NotGiven
                    try:
                        return cd.dec_value(d)
                    except Exception as exc:
                        self.logger.error("Decode: %s %r: %r", p, d, exc)
                        return NotGiven

            else:
                codec = self.codec

                def conv(p, d):
                    p  # noqa:B018
                    return d

            mon = await ex.enter_async_context(
                self.link.monitor(self.cf.dst, subtree=True, codec=codec)
            )
            task_status.started()
            ld = len(self.cf.dst)
            while True:
                try:
                    with anyio.fail_after(self.cf.get("timeout", 0.5)):
                        msg = await anext(mon)
                except TimeoutError:
                    break
                p = Path.build(msg.topic[ld:])
                res = conv(p, msg.data)
                if res is NotGiven:
                    continue
                await self.set_src(p, res, msg.meta)
            self.dst_is_current()

            async for msg in mon:
                if msg.meta is not None and msg.meta.origin == self.origin:
                    # mine, so skip
                    continue
                p = Path.build(msg.topic[ld:])
                if msg.data == b"":
                    res = NotGiven
                else:
                    res = conv(p, msg.data)
                    if res is NotGiven:
                        continue
                await self.set_src(p, res, msg.meta)

    async def set_dst(self, path: Path, data: Any, meta: MsgMeta, node: GateNode):  # noqa: D102
        meta = MsgMeta(origin=self.origin, timestamp=meta.timestamp)
        if data is NotGiven:
            await self.link.send(self.cf.dst + path, b"", retain=True, codec="noop", meta=meta)
        elif self.codecs is not None:
            try:
                vd = self.codec_vecs.search(path)
                cd = self.codecs.get(Path.build(vd.data["codec"]))
            except (ValueError, KeyError):
                self.logger.error("No codec: %s %r", path, data)
                return
            res = cd.enc_value(data)
            if isinstance(res, (str, bytes, bytearray)):
                await self.link.send(self.cf.dst + path, res, retain=True, codec="noop", meta=meta)
            else:
                self.logger.error("Bad codec: %s %r > %r", path, data, res)

        else:
            await self.link.send(
                self.cf.dst + path, data, retain=True, codec=self.codec, meta=meta
            )

        node.ext_meta = meta

    def is_update(self, node: GateNode, data: Any, meta: MsgMeta):
        """
        Test whether this is an update.

        @data is currently ignored.
        """
        data  # noqa:B018
        # if the old metadata match the new, it's not an update.
        try:
            if node.ext_meta.origin == meta.origin and node.ext_meta.timestamp == meta.timestamp:
                return False
        except (AttributeError, KeyError):
            pass
        return True

    def newer_dst(self, node):  # noqa: D102
        # If the external message has no metadata, it can't be from us,
        # thus assume it's newer.
        if not node.ext_meta:
            return True

        # If the internal and external metadata match, the message is from
        # us, so nothing to do.
        if self.origin == node.ext_meta.origin:
            return None

        # If the internal message has a copy of the outside metadata, it's
        # either unmodified or older. Test the data to be sure.
        if "gw" in node.meta:
            if node.meta["gw"] == node.ext_meta:
                return None if node.data_ == node.ext_data else True
            else:
                return True

        # Otherwise, if the external message is ours, it's old.
        if node.ext_meta.origin == self.origin:
            return False

        # if the timestamps are too close, there might be a problem.
        if abs(node.ext_meta.timestamp - node.meta.timestamp) < 0.1:
            return None

        # Otherwise use the message with the newer timestamp.
        return node.ext_meta.timestamp > node.meta.timestamp



# command line interface  # noqa: D100
from __future__ import annotations

import asyncclick as click

from moat.link.announce import announcing
from moat.link.client import Link


@click.group(short_help="Manage notifications.")  # pylint: disable=undefined-variable
@click.pass_context
async def cli(ctx):
    """
    Handle notifications.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg))


@cli.command()
@click.option("-b", "--backend", type=str, multiple=True, help="Restrict to this backend")
@click.pass_obj
async def run(obj, backend):
    """
    Forward notification messages.

    This command monitors the 'notify' subpath and forwards messages to
    your ntfy.sh instance.
    """
    from moat.link.notify import Notify  # noqa: PLC0415

    cfg = obj.cfg.link.notify
    if backend:
        cfg.backends = backend
    async with announcing(obj.conn) as ann:
        await Notify(cfg).run(obj.conn, evt=ann)



"""
Moat-Link is a library that enhances MQTT messaging (other backends are
possible) with structured data types, defined metadata (message source
and timestamp), and

The point of this library is to support a minimal unified pub/sub messaging
service.

Configuration looks like this:

    link:
      backend:
        driver: mqtt
        host: localhost
        port: 51883
        codec: std-cbor
      root: moat.org.example.test

"""

from __future__ import annotations

try:
    from importlib.metadata import version

    _version = version("moat.link")
    _version_tuple = tuple(int(x) for x in _version.split("."))

except Exception:  # pragma: no cover
    _version = "0.0.1"
    _version_tuple = (0, 0, 1)

protocol_version_min = 0
protocol_version = 0

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



"""
Connection and command helpers
"""

from __future__ import annotations

import anyio
from contextlib import asynccontextmanager

from moat.lib.cmd.anyio import run as run_stream

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.cmd import MsgSender
    from moat.lib.codec import Codec

    from collections.abc import AsyncGenerator


@asynccontextmanager
async def TCPConn(
    cmd: MsgSender | None,
    *a,
    codec: str | Codec = "std-cbor",
    debug: bool = False,
    logger=None,
    **kw,
) -> AsyncGenerator[..., MsgSender]:
    """
    Connection to a MoaT server.

    This encapsulates a TCP link to a remote side.

    Parameters:
    * the MsgHandler to connect to
    * all other arguments go to `anyio.connect_tcp`

    This is an async context manager. It forwards incoming commands to @cmd
    and yields a `MsgSender` that forwards local to the remote side.
    """
    async with (
        await anyio.connect_tcp(*a, **kw) as stream,
        run_stream(cmd, stream, codec=codec, debug=debug, logger=logger) as hdl,
    ):
        yield hdl


@asynccontextmanager
async def UnixConn(
    cmd: MsgSender | None,
    *a,
    codec: str | Codec = "std-cbor",
    debug: bool = False,
    logger=None,
    **kw,
) -> AsyncGenerator[..., MsgSender]:
    """
    Connection to a MoaT server.

    This encapsulates a Unix-domain client link.

    Parameters:
    * the MsgHandler to connect to
    * all other arguments go to `anyio.connect_unix`

    This is an async context manager. It forwards incoming commands to @cmd
    and yields a `MsgSender` that forwards local to the remote side.
    """
    async with (
        await anyio.connect_unix(*a, **kw) as stream,
        run_stream(cmd, stream, codec=codec, debug=debug, logger=logger) as hdl,
    ):
        yield hdl



"""
Notification package
"""

from __future__ import annotations

import anyio
import logging
import time
from abc import ABCMeta, abstractmethod
from contextlib import AsyncExitStack, asynccontextmanager

from moat.util import CtxObj, P, Path, as_service, attrdict

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.link.client import Link

    from collections.abc import AsyncIterator
    from typing import NoReturn, Self

__all__ = ["Notifier", "Notify", "get_backend"]

logger = logging.getLogger(__name__)


class Notify:
    """
    Notification runner.
    """

    def __init__(self, cfg):
        self.cfg = cfg

    async def run(self, link: Link, evt: anyio.Event | None = None):
        """
        Task that reads notifications from MoaT-Link and posts them.
        """
        self.link = link
        async with AsyncExitStack() as ex:
            backend = self.cfg.backend
            if isinstance(backend, str):
                backend = (backend,)

            self._backends = {}
            for name in backend:
                cfg = self.cfg.get(name, attrdict())
                try:
                    notifier = await ex.enter_async_context(get_backend(name, link, cfg))
                except Exception as exc:
                    logger.warning("Backend %r", name, exc_info=exc)
                else:
                    self._backends[name] = notifier

            if not self._backends:
                raise RuntimeError("No backend worked.")

            try:
                await self._run(evt)
            finally:
                with anyio.move_on_after(2, shield=True):
                    await self.send(
                        "error.notify", "Backend stopped", "The backend terminated.", prio="fatal"
                    )

    async def send(self, *a, **kw) -> None:
        """Send this notification to that topic."""
        bad = []
        dropped = []
        for name, b_e in list(self._backends.items()):
            try:
                await b_e.send(*a, **kw)
            except Exception as exc:
                logger.warning("Backend %r", name, exc_info=exc)
                bad.append(b_e)

        for name in bad:
            if self._backends.pop(name, None) is not None:
                dropped.append(name)
        if not self._backends:
            raise RuntimeError("All backends failed.")

        for name in dropped:
            await self.send(
                "error.notify",
                f"Backend {name} error",
                "The backend errored and was removed.",
                prio="error",
            )

    async def _run(self, evt) -> NoReturn:
        """
        A bridge that monitors the MoaT-Link notify subtree.
        """
        async with as_service(attrdict(debug=False)) as srv:
            try:
                await srv.tg.start(self._keepalive)

                async with self.link.d_watch(
                    self.cfg.path, subtree=True, meta=True, state=None
                ) as mon:
                    srv.set()
                    if evt is not None:
                        evt.set()
                    async for path, msg, meta in mon:
                        t = time.time()
                        if meta.timestamp < t - self.cfg.max_age:
                            continue
                        if isinstance(msg, dict):
                            if "title" not in msg:
                                msg["title"] = str(path)
                            await self.send(topic=path, **msg)
                        else:
                            await self.send(topic=path, title="?", msg=str(msg))

                    # not reached, loop doesn't terminate

            except Exception as exc:
                await self.send(
                    topic=P("error.notify"), msg=repr(exc), prio="error", title="Gateway failure"
                )
                raise

    async def _keepalive(self, *, task_status=anyio.TASK_STATUS_IGNORED) -> NoReturn:
        "Monitor the keepalive topic"
        bad = True
        link = self.link
        keep = self.cfg.keepalive
        ok_keep = keep.ok

        timeout = keep.get("timeout", link.cfg.timeout.ping.timeout)

        # The main host watcher publishes on the empty path
        async with link.d_watch(P("run.host"), state=None) as mon:
            task_status.started()
            mon = aiter(mon)  # noqa:PLW2901
            while True:
                if bad:
                    msg = await anext(mon)
                else:
                    try:
                        with anyio.fail_after(timeout):
                            msg = await anext(mon)
                    except TimeoutError:
                        msg = None

                if isinstance(msg, dict) and "id" in msg:
                    async with link.d_watch(P("run.ping.id") / msg["id"]) as mon2:
                        mon2 = aiter(mon2)  # noqa:PLW2901
                        while True:
                            try:
                                with anyio.fail_after(timeout):
                                    msg = await anext(mon2)
                                    if not msg.get("up", False):
                                        break
                                    if bad and "msg" in ok_keep:
                                        await self.send(topic="error.notify", **ok_keep)
                                        bad = False
                            except TimeoutError:
                                break

                await self.send(topic="error.notify", **keep)
                bad = True


def get_backend(name: str, link: Link, cfg: dict) -> Notify:
    """
    Fetch the backend named in the config and initialize it.
    """
    from importlib import import_module  # noqa: PLC0415

    if "." not in name:
        name = "moat.link.notify." + name
    return import_module(name).Notifier(link, cfg)


class Notifier(CtxObj, metaclass=ABCMeta):
    "Base class for notification backends"

    def __init__(self, link: Link, cfg: attrdict):
        self.cfg = cfg
        self.link = link

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        yield self

    @abstractmethod
    async def send(self, topic: str | Path, title: str, msg: str, **kw):
        """Send this notification to that topic."""



# command line interface  # noqa: D100
from __future__ import annotations

import anyio
import logging
from contextlib import nullcontext
from platform import uname

import asyncclick as click

from moat.util import as_service, attrdict, srepr
from moat.link.announce import announcing
from moat.link.client import Link
from moat.link.host import HostList, ServiceMon
from moat.util.broadcast import Broadcaster

logger = logging.getLogger(__name__)


@click.group(short_help="Manage host services.")  # pylint: disable=undefined-variable
@click.pass_context
async def cli(ctx):
    """
    Each server that's connected to moat-link should run a host service.

    This command manages that service.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if ctx.invoked_subcommand != "run":
        obj.conn = await ctx.with_async_resource(Link(cfg))


@cli.command()
@click.option("-m", "--main", is_flag=True, help="Main server flag (override)")
@click.option("-d", "--debug", is_flag=True, help="Debug?")
@click.pass_obj
async def run(obj, main, debug):
    """
    Host management background task.

    "moat link host run" should run on each MoaT-Link connected host.

    It provides keepalive-style ping messages and related services.
    """

    cfg = obj.cfg.link
    if not main:
        main = cfg.main == uname().node
    async with (
        Link(cfg) as link,
        as_service(attrdict(debug=debug, link=link)) as srv,
        announcing(link, host=not main, via=srv.evt, force=True),  # TODO: add service
        ServiceMon(cfg=cfg, link=link, debug=debug) if main else nullcontext(),
    ):
        srv.started()
        await anyio.sleep_forever()


@cli.command()
@click.option("-t", "--timeout", type=float, help="Stop after this many seconds.")
@click.pass_obj
async def list(obj, timeout):  # noqa: A001
    """
    Host list.

    "moat link host list" shows the hosts that are currently active.
    """

    with nullcontext() if timeout is None else anyio.move_on_after(timeout):
        async with HostList(link=obj.conn, cfg=obj.cfg.link, broadcaster=Broadcaster(10000)) as mq:
            async for h in mq:
                print("    UPD  ", h.id, h.state.name, srepr(h.data, bare=True))



"""
This module describes the MoaT data link.
"""

from __future__ import annotations

import anyio

# pylint: disable=missing-module-docstring
import logging
import sys
import time
from functools import partial

import asyncclick as click
from mqttproto import MQTTException

from moat.util import NotGiven, P, PathLongener, gen_ident, yprint
from moat.lib.codec import get_codec
from moat.link.backend import RawMessage, get_backend
from moat.util.path import set_root
from moat.util.times import humandelta, ts2iso

try:
    from mqttproto import MQTTPublishPacket
except ImportError:
    MQTTPublishPacket = NotGiven


logger = logging.getLogger(__name__)


@click.group
@click.pass_context
async def cli(ctx):
    """
    Commands that bypass the MoaT-Link server

    These access the MQTT backend directly.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]

    set_root(cfg)
    obj.conn = await ctx.with_async_resource(get_backend(cfg, name="mon_" + gen_ident()))


@cli.command()
@click.argument("file", type=click.Path(), nargs=1)
async def init(file):
    """Write an initial preload file.

    Usage: moat link raw init <node> <outfile>

    Writes an initial MoaT-Link data file.

    Using this command, followed by "moat kv server -l <outfile> <node>", is
    equivalent to running "moat kv server -i 'Initial data' <node>.
    """
    from moat.link.meta import MsgMeta  # noqa: PLC0415
    from moat.util.cbor import gen_start, gen_stop  # noqa: PLC0415
    from moat.util.msg import MsgWriter  # noqa: PLC0415

    meta = MsgMeta(origin="init")
    async with MsgWriter(path=file, codec="std-cbor") as f:
        await f(gen_start("Initial link data file", mode="init"))
        await f([0, P(":"), dict(type="MoaT-Link data"), *meta.dump()])
        await f(gen_stop(mode="log_end"))


@cli.command("test")
@click.pass_obj
async def test(obj):
    "Test"

    lock = anyio.Lock()
    cfg = obj.cfg.link

    async def check_root():
        try:
            with anyio.fail_after(1):
                async with back.monitor(cfg.root) as mon:
                    async for msg in mon:
                        async with lock:
                            print("Root dataset:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"No retained root dataset on {cfg.root} found.")

    async def check_server():
        try:
            with anyio.fail_after(1):
                async with back.monitor(cfg.root + P("run.service.main")) as mon:
                    async for msg in mon:
                        async with lock:
                            print("Server broadcast:")
                            yprint(msg.data)
                            print("---")
                            return
        except TimeoutError:
            print(f"No retained root dataset on {cfg.root} found.")

    async with get_backend(cfg) as back, back.connect(), anyio.create_task_group() as tg:
        tg.start_soon(check_root)
        tg.start_soon(check_server)


def _get_message(args):
    for m in args["msg"]:
        yield m
    for m in args["msg_eval"]:
        yield eval(m)  # pylint: disable=eval-used
    if args["msg_lines"]:
        with open(args["msg_lines"]) as f:  # pylint: disable=unspecified-encoding
            for line in f:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin_lines"]:
        for line in sys.stdin:
            if line:
                yield line.encode(encoding="utf-8")
    if args["msg_stdin"]:
        yield sys.stdin.buffer.read()
    if args["msg_stdin_eval"]:
        message = sys.stdin.read()
        yield eval(message)  # pylint: disable=eval-used


async def do_pub(client, args, cfg, codec):  # noqa: D103
    logger.info("%s Connected to broker", client.name)
    for k, v in args.items():
        if v is None or v is NotGiven:
            continue
        cfg[k] = v

    try:
        topic = args["topic"]
        retain = args["retain"]
        qos = args["qos"] or cfg["qos"]

        async with anyio.create_task_group() as tg:
            for message in _get_message(args):
                logger.info("%s Publishing to '%s'", client.name, topic)
                tg.start_soon(
                    partial(client.send, topic, message, qos=qos, retain=retain, codec=codec)
                )
        logger.info("%s Disconnected from broker", client.name)
    except KeyboardInterrupt:
        logger.info("%s Disconnected from broker", client.name)


@cli.command()
@click.option("-i", "--client_id", "--name", "name", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option("-r", "--retain", is_flag=True, help="Set the Retain flag")
@click.option("-t", "--topic", type=P, required=True, help="Message path")
@click.option("-m", "--msg", multiple=True, help="Message data (may be repeated)")
@click.option(
    "-M",
    "--msg-eval",
    multiple=True,
    help="Message data (Python, evaluated, may be repeated)",
)
@click.option(
    "-f",
    "--msg-lines",
    type=click.File("r"),
    help="File with messages (each line sent separately",
)
@click.option("-R", "--msg-stdin", is_flag=True, help="Single message from stdin")
@click.option(
    "-s",
    "--msg-stdin-lines",
    is_flag=True,
    help="Messages from stdin (each line sent separately",
)
@click.option(
    "-S",
    "--msg-stdin-eval",
    is_flag=True,
    help="Python code that evaluates to the message on stdin",
)
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.option("-c", "--codec", type=str, default="noop", help="codec to use (default: noop)")
@click.pass_obj
async def pub(obj, **args):
    """Publish one or more MoaT-Link messages.

    This command directly accesses the MQTT server: it works without a
    working MoaT-Link server.
    """
    if args["msg_stdin"] + args["msg_stdin_lines"] + args["msg_stdin_eval"] > 1:
        raise click.UsageError("You can only read from stdin once")
    cfg = obj.cfg.link
    codec = get_codec(args["codec"])

    if args["keep_alive"]:
        cfg["keep_alive"] = args["keep_alive"]

    await do_pub(obj.conn, args, cfg, codec)


async def run_kvsub(client, topic, lock, skip):
    """Monitor a MoaT-KV subtree"""

    if topic[-1] == "#":
        topic = topic.parent
        depth = -1
    else:
        depth = 0
    async with client.watch(
        topic,
        nchain=2,
        fetch=True,
        max_depth=depth,
        long_path=False,
    ) as res:
        pl = PathLongener(topic)
        async for r in res:
            async with lock:
                tm = time.time()
                atm = anyio.current_time()

                # if "value" in r:
                #    add_dates(r.value)
                pl(r)
                sk = False
                if "path" in r:
                    for s in skip:
                        if r.path.startswith(s):
                            sk = True
                            break
                if sk:
                    continue
                if r.get("state", "") == "uptodate":
                    continue
                del r["seq"]

                r["_prev"] = humandelta(atm - lock.tm, msec=6)
                r["_time"] = ts2iso(tm, msec=6)
                r["time"] = tm

                yprint(r)
                print("---")
                lock.tm = atm


async def do_sub(client, args, cfg):
    "handle subscriptions"
    lock = anyio.Lock()
    lock.tm = anyio.current_time()

    skip = args["skip"]
    try:
        async with anyio.create_task_group() as tg:
            for topic in args["topic"]:
                tg.start_soon(run_sub, client, topic, args, cfg.link, lock, skip)
            if args["kv_topic"]:
                from moat.kv.client import open_client as kv_client  # noqa: PLC0415

                async with kv_client(**cfg.kv) as kvc:
                    for topic in args["kv_topic"]:
                        tg.start_soon(run_kvsub, kvc, topic, lock, skip)

    except KeyboardInterrupt:
        pass
    except MQTTException as ce:
        logger.fatal("connection to '%s' failed: %r", args["uri"], ce)


async def run_sub(client, topic, args, cfg, lock, skip):
    "handle a single subscription"
    qos = args["qos"] or cfg["qos"]
    max_count = args["n_msg"]
    count = 0
    dcbor = get_codec("std-cbor")
    dmsgpack = get_codec("std-msgpack")

    async with client.monitor(topic, qos=qos, codec=args.get("codec", "noop")) as subscr:
        async for msg in subscr:
            sk = False
            for s in skip:
                if msg.topic.startswith(s):
                    sk = True
                    break
            if sk:
                continue
            async with lock:
                if args["yaml"]:
                    tm = time.time()
                    atm = anyio.current_time()

                    d = dict(
                        topic=msg.topic,
                        time=tm,
                        _prev=humandelta(atm - lock.tm, msec=6),
                        _time=ts2iso(tm, msec=6),
                    )

                    if isinstance(msg, RawMessage):
                        d["raw"] = msg.data
                        d["error"] = repr(msg.exc)
                    else:
                        d["data"] = msg.data
                    if msg.meta is not None:
                        d["meta"] = msg.meta.repr()

                    try:
                        d["data_cbor"] = dcbor.decode(msg.data)
                    except Exception:  # noqa:S110
                        # not a valid CBOR message, which is OK
                        pass

                    try:
                        d["data_msgpack"] = dmsgpack.decode(msg.data)
                    except Exception:  # noqa:S110
                        # not a valid msgpack message, which is OK
                        pass

                    flags = ""
                    if msg.retain:
                        flags += "R"
                    # if msg.qos > 0:
                    #    flags += f"Q{int(msg.qos)}"
                    if flags:
                        d["_flags"] = flags

                    lock.tm = atm
                    yprint(d)
                    print("---")
                else:
                    if isinstance(msg, RawMessage):
                        print(msg.topic, "*", msg.data, repr(msg.exc), sep="\t")
                    else:
                        print(msg.topic, msg.data, sep="\t")
            count += 1
            if max_count and count >= max_count:
                break


@cli.command()
@click.option("-i", "--client_id", "--name", "name", help="string to use as client ID")
@click.option("-q", "--qos", type=click.IntRange(0, 2), help="Quality of service to use (0-2)")
@click.option("-y", "--yaml", is_flag=True, help="Print output as YAML stream")
@click.option(
    "-t",
    "--topic",
    multiple=True,
    type=P,
    help="Message topic, dot-separated (can be used more than once)",
)
@click.option(
    "-T",
    "--kv-topic",
    multiple=True,
    type=P,
    help="Message topic from MoaT-KV legacy, dot-separated (can be used more than once)",
)
@click.option("-n", "--n_msg", type=int, default=0, help="Number of messages to read (per topic)")
@click.option("-k", "--keep-alive", type=float, help="Keep-alive timeout (seconds)")
@click.option("-s", "--skip", type=P, multiple=True, help="Skip this path prefix")
@click.pass_obj
async def sub(obj, **args):
    """
    Monitor one or more MoaT-Link topics.

    This command directly accesses the MQTT server: it works without a
    working MoaT-Link server.
    """
    cfg = obj.cfg

    if args["keep_alive"]:
        cfg.link["keep_alive"] = args["keep_alive"]

    await do_sub(obj.conn, args, cfg)



"""
MoaT-Link services announce themselves with defined names.
"""

from __future__ import annotations

import anyio
import logging
import os
import platform
import warnings
import weakref
from anyio.abc import TaskStatus
from contextlib import asynccontextmanager

from moat.util import NotGiven, P, Path, attrdict, ensure_cfg, gen_ident
from moat.util import as_service as _as_service

from .client import Link
from .exceptions import ServiceNotFound, ServiceNotStarted, ServiceSupplanted

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.cmd.base import MsgSender
    from moat.util.path import PathElem

    from .client import LinkSender

    from typing import Any

__all__ = ["announcing"]

logger = logging.getLogger(__name__)

# TODO when to warn vs. when to log a warning


async def get_service_path(host: Path | str | bool):
    path: list[PathElem]
    match host:
        case True:
            path = [platform.node()]
        case False:
            path = []
        case Path():
            path = list(host.raw)
        case str():
            path = [host]
        case _:
            raise TypeError("Host must be bool/path/str, not {host !r}.")

    if (fake := os.environ.get("MT_FAKE_SERVICE", None)) is not None:
        path.extend(P(fake))
        return path

    if int(os.environ.get("SYSTEMD_EXEC_PID", "0")) == os.getpid():
        async with await anyio.Path("/proc/self/cgroup").open("r") as cgf:
            async for cg in cgf:
                logger.info("Control Group: %s", cg)
                for cge in cg.strip().split("/"):
                    if cge.endswith(".service"):
                        cge = cge[:-8]  # noqa:PLW2901
                        hi = cge.split("@", 1)
                        if hi[0] == "user":
                            continue  # ignore
                        if len(hi) != 1 or hi[0] != "moat-link-host":
                            for h in hi:
                                path.extend(h.split("-"))
                        logger.info("Control Path: %s", path)
                        return path
    raise ServiceNotFound


class SetReady(TaskStatus):
    """A fake event/TaskStatus to signal readiness"""

    value: Any = None

    def __init__(self, ann: anyio.Event, host, name):
        self.evt = anyio.Event()
        self.ann = ann
        self.host = host
        self.name = name

    def set(self):  # pylint:disable=missing-function-docstring
        "Set the event."
        self.evt.set()

    async def announce(self, value: Any = None):
        """
        Store the value if any, set the event, wait for the announcement to
        be transmitted.
        """
        self.started(value)
        await self.ann.wait()

    def started(self, value: Any = None):
        """
        Mock task_status.started. A value, if any, is added to
        the Host message as its ``val`` item.
        """
        self.value = value
        self.set()

    def __del__(self):
        """
        Complain if the SR was unused.
        """
        try:
            if not self.evt.is_set():
                warnings.warn(
                    f"event {self.name} on {self.host} freed before announcing",
                    ServiceNotStarted,
                )
        except Exception:  # noqa:S110
            pass

    def warn(self, s: str):
        logger.warning("Service %s on %s %s", self.name, self.host, s)


class _csr:
    # This object caches a SetReady until it is called.
    # The idea is to be able to atomically yield *and* delete our reference to a SR.

    def __init__(self, sr: SetReady):
        self.sr = sr

    def __call__(self):
        sr = self.sr
        del self.sr
        return sr

    def __del__(self):
        try:
            # Don't worry if any of these no longer work.
            self.sr.evt.set()
        except Exception:  # noqa:S110
            pass


@asynccontextmanager
async def announcing(
    link: LinkSender,
    name: Path | None = None,
    *,
    host: Path | str | bool = True,
    force: bool = False,
    service: MsgSender | None = None,
    via: anyio.Event | None = None,
):
    """
    This async context manager broadcasts the availability of a named service.

    Arguments:
        link: the link to use.
        host: a way to override the discovered service name.
        name: additional elements on the announcement
        service: Service to delegate to
        force: Flag to override an existing server

    The CM yields a SetReady object. You must call one of its ``.set``,
    ``annonce`` or ``started`` methods to start the announcement.

    Services are declared by a host prefix and a named path.

    ``host`` can be
    * ``True``: Use the hostname of the system the service runs on.
      This is the default.
    * a string: replace the hostname of the system the service runs on.
      Used for services that should run once in an installation.
    * ``False``: Don't add a hostname. This should not normally be done.

    If ``name`` is `None`, it will be auto-discovered from
    the current systemd unit, by selecting service entries in
    ``/proc/self/cgroups``. It is an error for the result to be empty.
    This requires the ``SYSTEMD_EXEC_PID`` environment variable to
    contain the PID of the current process.

    A :class:`ServiceSupplanted` exception will be raised if/when there is
    a duplicate of the named service on the link (or one appears).

    The ``force`` flag is intended for replacing an existing service. It
    does not ignore a subsequent announcement by some other MoaT-Link
    client.
    """

    if service is not None:
        path = Path(gen_ident(12))

    async def run_announce(srv: Path, sr: SetReady, rdy: anyio.Event) -> None:
        # Send our announcement when ready
        evt = sr.evt
        csr = weakref.ref(sr)
        del sr

        try:
            await evt.wait()
        except BaseException:
            # If we're cancelled (or whatever), set the event. The SR will
            # still be referenced at this point, because the frame that
            # called "announcing" is somewhere in the exception chain, thus
            # this is guaranteed to run before SR.__del__.
            evt.set()
            raise

        data: dict[str, Any] = {"id": link.id}
        if service:
            data["path"] = path

        # if it's already gone, don't worry.
        sr = csr()
        if sr is not None and sr.value is not None:
            data["val"] = sr.value
        await link.d_set(srv, data, retain=True)
        await link.i_sync()
        rdy.set()

    async def monitor_service(srv: Path):
        # Watch for others
        async with link.d_watch(srv, state=False if force else NotGiven) as mon:
            async for msg in mon:
                if msg["id"] != link.id and await link.i_checkid(msg["id"]):
                    raise ServiceSupplanted(srv)

    async def wait_sr(sr: SetReady):
        csr = weakref.ref(sr)
        del sr
        await anyio.sleep(5)
        try:
            csr().warn("did not start")
        except Exception:  # noqa:S110
            pass

    async def _delegate(path, service, *, task_status: anyio.TASK_STATUS_IGNORED):
        with link.link.delegate(path, service):
            task_status.started()
            await anyio.sleep_forever()

    async def wait_evt(evt: anyio.Event):
        await via.wait()
        evt.set()

    async with anyio.create_task_group() as tg:
        try:
            rdy = anyio.Event()
            tm = anyio.current_time()
            sr = SetReady(rdy, host, name)
            csr = _csr(sr)
            evt = sr.evt
            try:
                srv = P("run.host") + ((await get_service_path(host)) if name is None else name)
            except ServiceNotFound:
                logger.warning("Not running in a service CGroup. Host not set.")
                yield None
                return
            if service:
                await tg.start(_delegate, path, service)
            tg.start_soon(monitor_service, srv)
            tg.start_soon(run_announce, srv, sr, rdy)
            tg.start_soon(wait_sr, sr)
            if via:
                tg.start_soon(wait_evt, evt)
            del sr

            await link.i_sync()

            yield csr()
            if not evt.is_set():
                evt.set()
                logger.warning("Service %s on %s did not start", name, host)

        except ServiceSupplanted:
            logger.warning("Service %s on %s already exists", name, host)
            raise
        except BaseException:
            if anyio.current_time() - tm > 5 and not evt.is_set():
                logger.warning("Service %s on %s did not starting", name, host)
            raise
        finally:
            tg.cancel_scope.cancel()


@asynccontextmanager
async def as_service(obj: attrdict | None = None):
    """
    This is a replacement for the legacy :func:`moat.util.as_service`
    helper that also registers the named service with MoaT-Link.
    """
    if obj is None:
        obj = attrdict()
    obj.setdefault("debug", False)
    cfg = ensure_cfg("moat.link", obj.cfg)

    async with (
        _as_service(obj) as mon,
        Link(cfg.link, common=True) as mon.link,
        announcing(mon.link, via=mon.evt),
    ):
        yield mon



# command line interface  # noqa: D100
from __future__ import annotations

import asyncclick as click

from moat.util import NotGiven, P, attr_args, process_args, yprint
from moat.link.client import Link


@click.command(short_help="Send a command")
@click.option("-S", "--stream", is_flag=True, help="Read a stream")
@click.option("-R", "--raw", is_flag=True, help="Show raw message data")
@click.argument("path", type=P, nargs=1)
@click.pass_context
@attr_args
async def cli(ctx, path, stream, raw, **kw):
    """
    Send a command to the server.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    if obj.get("port", None) is not None:
        cfg.client.port = obj.port

    async with Link(cfg) as conn:
        val = process_args(NotGiven, **kw)
        kw = {}
        if isinstance(val, list):
            args = val
        elif isinstance(val, dict):
            args = []
            for k, v in list(val.items()):
                if k.startswith("_"):
                    try:
                        k = int(k[1:])  # noqa:PLW2901
                    except ValueError:
                        pass
                    else:
                        if k >= len(args):
                            args += [NotGiven] * (k - len(args) + 1)
                        args[k] = v
                        continue
                kw[k] = v
        elif val is NotGiven:
            args = []
        else:
            args = [val]

        if stream:
            async with conn.cmd(path, *args, **kw).stream_in() as res:
                yprint(rep_(res, raw), stream=obj.stdout)
                print("---", file=obj.stdout)
                async for msg in res:
                    yprint(rep_(msg, raw), stream=obj.stdout)
                    print("---", file=obj.stdout)
        else:
            res = await conn.cmd(path, *args, **kw)
        yprint(rep_(res, raw), stream=obj.stdout)


def rep_(res, raw=False):  # noqa: D103
    if raw:
        return res.args + [res.kw]
    if not res.args:
        return res.kw
    if res.kw:
        return res.args + [res.kw]
    if len(res.args) == 1:
        return res.args[0]
    return res.args



"""
The MoaT-Link client
"""

from __future__ import annotations

import anyio
import logging
import os
import platform
import sys
import time
from contextlib import asynccontextmanager, nullcontext, suppress
from contextvars import ContextVar
from functools import partial
from traceback import format_exception

import outcome
from attrs import define, field
from mqttproto import QoS, RetainHandling

from moat.util import (
    CtxObj,
    NotGiven,
    P,
    Path,
    PathLongener,
    Root,
    ValueEvent,
    attrdict,
    ctx_as,
    gen_ident,
    srepr,
    timed_ctx,
    ungroup,
)
from moat.lib.cmd.base import Caller, MsgSender
from moat.util.random import al_unique

from .auth import AnonAuth, TokenAuth
from .common import CmdCommon
from .conn import TCPConn, UnixConn
from .exceptions import AuthError, ClientCancelledError
from .hello import Hello
from .meta import MsgMeta
from .node import Node

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from moat.lib.cmd.base import MsgHandler
    from moat.lib.cmd.msg import Msg
    from moat.link.node.codec import CodecNode

    from .backend import Message
    from .schema import Data
    from .schema import SchemaName as S

    from collections.abc import AsyncIterator, Awaitable
    from typing import Any, Literal


class _Requeue(Exception):
    pass


__all__ = ["BasicLink", "Link", "LinkCommon", "get_link"]

_the_link = ContextVar("_the_link")

local_addrs = {"localhost", "127.0.0.1", "::1"}


def get_link() -> Link | None:
    """
    Return the global Link instance, if one exists.
    """
    return _the_link.get(None)


class TS(anyio.abc.TaskStatus):
    "A wrapper to TaskStatus that swallows successive calls"

    def __init__(self, ts):
        self.ts = ts

    def started(self, data=None):
        "call .started once"
        self.ts.started(data)
        self.ts = anyio.TASK_STATUS_IGNORED


class BasicCmd:
    """
    A simple command that doesn't require streaming.

    The command can thus be repeated if a connection dies while it's
    running.
    """

    def __init__(self, a, kw):
        self.a = a
        self.kw = kw
        self._result = None
        self._evt = anyio.Event()

    async def run(self, link):
        "run the command"
        try:
            res = await link.cmd(*self.a, **self.kw)
        except anyio.get_cancelled_exc_class():
            raise
        except Exception as exc:
            self._result = outcome.Error(exc)
        else:
            self._result = outcome.Value(res)
        finally:
            self._evt.set()

    @property
    async def result(self):
        "retreve the result"
        await self._evt.wait()
        if self._result is None:
            raise _Requeue
        return self._result.unwrap()


class LinkCommon(CmdCommon):
    """
    Common code for links to MQTT only (BasicLink) and to a MoaT-Link server (Link).
    """

    protocol_version: int = -1
    name: str
    server_name: str
    is_server: bool = False

    def __init__(self, cfg, name: str | None = None):
        self.cfg = cfg
        self._id = gen_ident(12, alphabet=al_unique)
        if name is not None:
            self.is_server = True
        else:
            name = cfg.get("client_id")
            if name is None:
                name = "_" + self._id
        self.name = name

        self._cmdq_w, self._cmdq_r = anyio.create_memory_object_stream(5)
        self.logger = logging.getLogger(f"moat.link.client.{name}")
        self.sender = MsgSender(self)

    def handle(self, msg, rpath, *add) -> Awaitable[Any]:
        """
        Message handler that intercepts incoming commands
        while authorization has not completed
        """
        if self._hello is not None and self._hello.auth_data is None:
            return self._hello.handle(msg, rpath, *add)

        return super().handle(msg, rpath, *add)

    @property
    def id(self):  # noqa:D102
        return self._id

    async def _connected_port(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        async with self._connect_one(self._port) as hdl:
            task_status.started(hdl)
            await anyio.sleep_forever()

    @asynccontextmanager
    async def _connect_one(self, remote: dict | str, data: dict | None = None) -> MsgHandler:
        auth_out = []
        if isinstance(remote, dict):
            with suppress(KeyError):
                auth_out.append(TokenAuth(data["auth"]["token"]))
            conn_ = TCPConn(
                self, remote_host=remote["host"], remote_port=remote["port"], logger=self.logger
            )
        else:
            conn_ = UnixConn(self, path=remote, logger=self.logger.debug)

        auth_out.append(AnonAuth())
        self._hello = Hello(me=self.name, me_server=self.is_server, auth_out=auth_out)
        yielded = False

        async with conn_ as conn:
            if self.logger.isEnabledFor(logging.INFO):
                self.logger.info("Connection %s to %s", self.name, srepr(remote))
            handler = MsgSender(conn)
            if (res := await self._hello.run(handler)) is False:
                raise AuthError("Initial handshake failed")

            if isinstance(res, dict):
                name = res.pop("name", None)
                if name is not None:
                    self.name = name
                if res:
                    self.logger.warning("Unknown auth reply: %r", res)

            self.name = self._hello.me
            self.server_name = self._hello.them
            self.protocol_version = self._hello.protocol_version
            self._hello = None  # done with that

            yielded = True
            yield handler

        if not yielded:
            raise EOFError


class ClientCaller(Caller):  # nqa:D102
    def __init__(self, sender, *a, **kw):
        self._link = sender._link  # noqa:SLF001
        super().__init__(sender, *a, **kw)

    @asynccontextmanager
    async def _ctx(self):
        await self._link.get_link()
        async with super()._ctx() as res:
            yield res

    async def _call(self):
        "helper for __await__ that calls the remote handler"
        link = await self._link.get_link()
        cmd, a, kw = self.data
        return await link.root._sender.cmd(cmd, *a, **kw)  # noqa:SLF001


class LinkSender(MsgSender):
    """
    This is the client-side front-end to a "standard" MoaT-Link connection.

    It contains various somewhat-high-level helpers. Most are named like
    the corresponding backend commands, with an underscore replacing the
    dot.
    """

    Caller_ = ClientCaller

    _codec_tree: CodecNode | None = None

    def __init__(self, link: LinkCommon):
        self._link = link

    @property
    def root(self):
        return self

    @property
    def link(self):
        return self._link

    @property
    def name(self) -> str:
        return self._link.name

    @property
    def cfg(self) -> attrdict:
        return self._link.cfg

    @property
    def id(self) -> str:
        return self._link.id

    async def handle(self, msg: Msg, rcmd: list) -> Awaitable[None]:
        """
        Standard handler, forwards to the remote side.
        """
        srv = await self._link.get_link()
        await srv.handle(msg, rcmd)

    def find_handler(self, path, cmd: bool = False) -> tuple[MsgHandler, Path]:
        """
        Standard sub-dispatcher redirector, no-op.
        """
        cmd  # noqa:B018
        return self, path

    @overload
    def d_get(self, path: Path, meta: Literal[True]) -> tuple[Any, MsgMeta]: ...

    @overload
    def d_get(self, path: Path) -> Any: ...

    async def d_get(self, path: Path, meta: bool = False) -> tuple[Any, MsgMeta]:
        """
        Data retrieval. Calls the server's ``d.get`` method.

        Returns a data+metadata tuple if @meta is True, otherwise just the
        data.
        """
        if len(path) and isinstance(path[0], Path):
            raise ValueError("Don't use a root-prefixed path here.")

        res = await self.d.get(path)
        if not meta:
            return res[0]
        return res[0], MsgMeta.restore(res[1:])

    @overload
    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        t: float | None = None,
        meta: Literal[True] = True,
        retain: bool | None = None,
    ) -> tuple[Any, MsgMeta]: ...

    @overload
    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        t: float | None = None,
        meta: Literal[False] = False,
        retain: bool | None = None,
    ) -> None: ...

    async def d_set(
        self,
        path: Path,
        data: Any = NotGiven,
        meta: MsgMeta | None = None,
        t: float | None = None,
        with_prev: bool = False,
        retain: bool | None = None,
    ) -> None | tuple[Any, MsgMeta]:
        """
        Data update.

        If a timestamp is passed in @t or the old value+metadata is
        requested via @with_prev, goes through the server. Otherwise posts to
        MQTT directly.
        """
        if path and isinstance(path[0], Path):
            raise ValueError("Don't use a root-prefixed path here.")

        if meta is None:
            meta = MsgMeta(origin=self._link.name)
        if t is None and not with_prev:
            if retain is None:
                retain = len(path) == 0 or path[0] != "run"
            await self.send(Root.get() + path, data=data, meta=meta, retain=retain)
            return
        tt = {} if t is None else {"t": t}
        res = await self.d.set(path, data, meta, **tt)
        if not with_prev:
            return res[0]
        meta = MsgMeta.restore(res[1:]) if len(res) > 1 else None
        return res[0], meta

    @asynccontextmanager
    async def d_walk(
        self,
        path: Path,
        meta: bool = False,
        min_ts: float = 0,
        min_depth: int = 0,
        max_depth: int = 255,
    ) -> AsyncIterator[tuple[str, Any, MsgMeta]]:
        """
        Fetch a (limited) subtree.
        """
        args = [path, min_ts, min_depth, max_depth]
        if args[-1] == 255:
            args.pop()
            if args[-1] == 0:
                args.pop()
                if args[-1] == 0:
                    args.pop()

        async with self.d.walk(*args).stream_in() as mon:
            yield Walker(mon, meta=meta)

    # No use marking as Any|None if @mark is set
    @overload
    def d_watch(
        self,
        path: Path,
        mark: bool = False,
        meta: Literal[False] = False,
        subtree: Literal[False] = False,
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[Any]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False] = False,
        meta: Literal[True] = True,
        subtree: Literal[False] = False,
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False],
        meta: Literal[True],
        subtree: Literal[True],
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Path, Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[False] = False,
        meta: Literal[False] = False,
        subtree: Literal[True] = True,
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Path, Any]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[True],
        subtree: Literal[False] = False,
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[True],
        subtree: Literal[True],
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Path, Any, MsgMeta]]]: ...

    @overload
    def d_watch(
        self,
        path: Path,
        mark: Literal[True],
        meta: Literal[False] = False,
        subtree: Literal[True] = True,
        state: bool | Literal[NotGiven] | None = None,
    ) -> AbstractAsyncContextManager[AsyncIterator[None | tuple[Path, Any]]]: ...

    def d_watch(
        self,
        path: Path,
        meta: bool = False,
        subtree: bool = False,
        state: bool | None | NotGiven = None,
        age: float | None = None,
        mark: bool = False,
        min_length: int | None = None,
        max_length: int | None = None,
        cls: type = Node,
    ) -> AbstractAsyncContextManager[AsyncIterator[tuple[Path, Any, MsgMeta]]]:
        """
        Monitor a node or subtree.

        @path: what to examine
        @meta: flag whether to return metadata too
        @subtree: flag whether to watch a subtree, not just this node
        @mark: yield `None` when the initial state has been transmitted
        @state: send the current state (True), updates (False), both with
                current data from the server (None), or both via MQTT (NotGiven).
        @age: cutoff this many seconds ago. Older entries are skipped.
        @cls: type of root node (default `Node`)

        This method returns an async context manager which yields an async iterator.
        The iterator yields node data if neither @subtree nor @meta is set.
        Otherwise it yields tuples. The first item is the path if @subtree
        is set; the last item is the metadata if @meta is set.

        ``state=NotGiven`` does not work with @mark=True, or when the same
        subscription is already active somewhere else.
        """
        return Watcher(self, path, meta, subtree, state, age, mark, cls, min_length, max_length)

    async def e_exc(self, path: Path, exc: Exception, **kw):
        """
        Report an exception, with backtrace.
        """
        exc = ungroup.one(exc)

        kw["_bt"] = format_exception(exc)
        kw["_exc"] = exc

        await self.e.exc(path, repr(exc), **kw)

    async def e_info(self, path: Path, txt: str, **kw):
        """
        Report a problem that's not an exception.
        """
        await self.e.info(path, txt, **kw)

    async def e_ack(self, path: Path, **kw):
        """
        Report that $USER has been notified of a problem.

        @ack can be True (don't notify any more), False (notify next
        time it happens), or a timestamp (notify if it happens again after
        $time).
        """
        await self.e.ack(path, **kw)

    async def e_ok(self, path: Path, **kw):
        """
        Report that something is working.
        """
        await self.e.ok(path, **kw)

    @asynccontextmanager
    async def e_wrap(self, path: Path, **kw):
        """
        An error-handling context manager. It opens a stream to the server
        and reports either success or failure over it.

        Streamed messages can be sent and will be stored as auxiliary
        information if the command should fail.

        This wrapper does *not* swallow the exception.
        """
        async with self.e.mon(path, **kw) as mon:
            yield mon
            await mon.send(True)

    def monitor(self, path: Path, *a, **kw) -> AbstractAsyncContextManager[AsyncIterator[Message]]:
        """
        Watch this path.

        This call forwards directly to the back-end.
        See `moat.link.backend.Backend.monitor` for call details.

        The caller is responsible for prefixing the Root path, if applicable.
        """
        return self._link.backend.monitor(path, *a, **kw)

    def send(self, path: Path, *a, **kw) -> Awaitable[None]:
        """
        Send to this path.

        This call forwards directly to the back-end.
        See `moat.link.backend.Backend.send` for call details

        The caller is responsible for prefixing the Root path, if applicable.
        """
        return self._link.backend.send(path, *a, **kw)

    async def i_sync(self):
        """
        This call tries to ensure that the server has processed all
        incoming MQTT messages.

        It does this by requesting a timestamp from the server, publishing
        the timestamp to MQTT, and then waiting for the server to
        acknowledge having received (at least) this number.

        This code assumes that MQTT messages are delivered in the order
        they are submitted. Depending on the server's multithreading setup
        this may or may not be true, but in practice it should be
        sufficiently true-ish to work out.

        It also assumes that if there's more than one server, they all have
        mostly-accurate time.
        """
        (st,) = await self.cmd(P("i.stamp"))
        await self.send(P(":R.run.service.main.stamp") / self._link.server_name, st, retain=False)
        await self.cmd(P("i.sync"), st)

    async def i_checkid(self, id: str) -> bool:
        """
        Check whether the given client ID is still valid.
        """
        (st,) = await self.cmd(P("i.checkid"), id)
        return st

    async def _get_tree(self, path, *, task_status, **kw):
        async with self.d_watch(path, **kw) as w:
            task_status.started(await w.get_node())
            await anyio.sleep_forever()

    async def get_codec_tree(self):
        """
        Returns a common subtree for codecs.
        """

        if self._codec_tree is None:
            self._codec_tree = evt = anyio.Event()

            from moat.link.node.codec import CodecNode  # noqa: PLC0415

            self._codec_tree = await self._link.tg.start(
                partial(
                    self._get_tree, P("codec"), subtree=True, state=None, meta=False, cls=CodecNode
                )
            )
            evt.set()

        elif isinstance(self._codec_tree, anyio.abc.Event):
            await self._codec_tree.wait()

        return self._codec_tree

    async def get_service(self, srv: Path) -> MsgHandler:
        res = await self.d_get(P("run.host") + srv)
        return self.sub_at(P("cl") / res["id"] + res["path"])


class Link(LinkCommon, CtxObj):
    """
    This class combines the back-end link with a connection to a MoaT-Link server.

    If "common" is set, create global state.
    """

    _server: ValueEvent = None
    _uptodate: bool = False
    _hello: Hello = None
    current_server: MsgSender = None
    _server_up: anyio.Event
    _last_link: Msg | None = None
    _last_link_seen: anyio.Event
    _port: str | None = None
    _state: str = "init"
    _common: bool = False

    def __new__(cls, cfg, name: str | None = None, common: bool = False):  # noqa:D102
        cfg, name  # noqa:B018
        if common and (link := _the_link.get(NotGiven)) is not NotGiven:
            return link
        return super().__new__(cls)

    def __init__(self, cfg, name: str | None = None, common: bool = False):
        if _the_link.get(NotGiven) is self:
            return
        super().__init__(cfg, name=name)
        self._retry_msgs: set[BasicCmd] = set()
        self._server_up = anyio.Event()
        self._state_change = anyio.Event()
        self._common = common
        with suppress(AttributeError):
            self._port = self.cfg.client.port

    async def set_state(self, state: str):
        """
        Set the state string for our ping message
        """
        if state != self._state:
            self._state = state
            self._state_change.set()

    async def get_link(self):
        """
        This method refreshes the link to the server if it happens to be down.
        """
        while self.current_server is None:
            await self._server_up.wait()
        return self.current_server

    @property
    def _ping_path(self):
        return P("run.ping.id") / self._id

    @property
    def _id_path(self):
        return P("run.id") / self._id

    async def _send_id(self):
        await self.sdr.d_set(
            self._id_path,
            data=dict(
                host=platform.node(),
                pid=os.getpid(),
                argv=sys.argv,
            ),
            retain=True,
        )

    async def _send_ping(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        Periodially (or when our state changes) send a Ping.

        This also sends our ID message, after the first ping,
        but only if the task status isn't ignored.
        """
        path = Root.get() + self._ping_path
        while True:
            await self.backend.send(
                path,
                data=dict(
                    up=True,
                    state=self._state,
                ),
                retain=False,
                meta=False,
            )

            if task_status is not anyio.TASK_STATUS_IGNORED:
                # send initial run.id message, *after* the ping
                await self._send_id()
                task_status.started()
                task_status = anyio.TASK_STATUS_IGNORED

            if self._state == "init":
                self._state = "auto"

            with anyio.move_on_after(self.cfg.timeout.ping.every):
                await self._state_change.wait()
                self._state_change = anyio.Event()

    async def _monitor_ping(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        path = Root.get() + self._ping_path
        async with self.backend.monitor(path) as mon:
            task_status.started()
            async for msg in mon:
                if msg.data is NotGiven:
                    raise ClientCancelledError(msg.meta)
                # anything else shall be ignored

    @asynccontextmanager
    async def _ctx(self):
        from .backend import get_backend  # noqa: PLC0415

        will = attrdict(
            data=dict(up=False, state="will"),
            topic=P(":R") + self._ping_path,
            retain=False,
            qos=QoS.AT_LEAST_ONCE,
        )
        async with (
            ctx_as(Root, self.cfg["root"]),
            get_backend(self.cfg, name=self.name, will=will) as self.backend,
            anyio.create_task_group() as self.tg,
        ):
            if self._port is not None:
                sdr = await self.tg.start(self._connected_port)
            else:
                if self.cfg.client.init_timeout:
                    # connect to the main server
                    await self.tg.start(self._run_server_link)
                sdr = LinkSender(self)
            sdr.add_sub("cl")
            sdr.add_sub("d")
            sdr.add_sub("e")
            sdr.add_sub("i")
            try:
                self.sdr = sdr
                await self.tg.start(self._monitor_ping)
                await self.tg.start(self._send_ping)

                with ctx_as(_the_link, self) if self._common else nullcontext():
                    yield sdr
            finally:
                del self.sdr
            await self.backend.send(
                Root.get() + self._ping_path,
                data=dict(up=False, state="closed"),
                retain=False,
                meta=False,
            )
            self.tg.cancel_scope.cancel()

    def cancel(self):
        "Stop me"
        self.tg.cancel_scope.cancel()

    async def _run_server_link(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        """
        This is the manager task for the server link channel.
        It starts a server connection (and tries to keep it alive).
        """
        task_status = TS(task_status)

        with anyio.fail_after(self.cfg.client.init_timeout):
            srv = await self.tg.start(self._read_server_link)

        tm = self.cfg.timeout.connect
        timeout = tm.initial
        while True:
            try:
                await self._connect_server(srv, task_status=task_status)
            except Exception as exc:
                await self.backend.send_error(
                    P("run.service.main.server") / srv.meta.origin,
                    data=srv,
                    exc=exc,
                )
            finally:
                self.current_server = None
                if self._server_up.is_set():
                    self._server_up = anyio.Event()

                # TODO test if tasks from the remote side already get cancelled
                # TODO and if not, do that

            # Back off to either a new server announcement or the exponential timeout
            try:
                with anyio.fail_after(timeout):
                    await self._last_link_seen.wait()
            except TimeoutError:
                # try the last-reported connection again
                timeout = min(timeout * tm.factor, tm.max)
            else:
                # immediately use the new data
                srv = self._last_link
                self._last_link_seen = anyio.Event()

    async def _read_server_link(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        # Background process that monitors the server link channel

        self._last_link_seen = anyio.Event()
        async with self.backend.monitor(
            P(":R.run.service.main.conn"),
            retain_handling=RetainHandling.SEND_RETAINED,
        ) as mon:
            async for msg in mon:
                self._last_link = msg
                if task_status is None:
                    self._last_link_seen.set()
                else:
                    task_status.started(msg)
                    task_status = None

    async def _connect_run(self, rem, *, task_status=anyio.TASK_STATUS_IGNORED):
        # We're connected.
        self.current_server = rem
        self._server_up.set()

        async def run_(cmd):
            await cmd.run(rem)

        async with anyio.create_task_group() as tg:
            for msg in self._retry_msgs:
                tg.start_soon(run_, msg)
            task_status.started()
            async for msg in self._cmdq_r:
                tg.start_soon(run_, msg)

    async def cmd(self, *a, _idem=False, **kw):
        """
        Queue and run a simple command.

        If @_idem is False, the command will error out if the connection
        dies while it's running. Otherwise (the default) it may be repeated.
        """
        cmd_ = BasicCmd(a, kw)
        try:
            self._retry_msgs.add(cmd_)
            while True:
                while self.current_server is None:
                    await self._server_up.wait()

                try:
                    with ungroup:
                        await self._cmdq_w.send(cmd_)
                        return await cmd_.result
                except _Requeue:
                    if _idem:
                        raise EOFError from None

        finally:
            self._retry_msgs.discard(cmd_)

    async def _connect_server(
        self,
        srv: Message[Data[S.run.service.main.conn]],
        *,
        task_status=anyio.TASK_STATUS_IGNORED,
    ):
        task_status = TS(task_status)

        # Backend connection
        link = srv.data["link"]
        if isinstance(link, dict):
            link = (link,)

        local_ok = srv.data.get("node", "localhost")

        for remote in link:
            if local_ok:
                pass
            elif "host" not in remote:
                continue
            elif remote["host"] in local_addrs:
                continue
            try:
                async with timed_ctx(
                    self.cfg.client.init_timeout, self._connect_one(remote, srv.data)
                ) as rem:
                    await self._connect_run(rem, task_status=task_status)
            except OSError as exc:
                self.logger.warning("Link failed: %r (%r)", remote, exc)
            except Exception as exc:
                self.logger.warning("Link failed: %r", remote, exc_info=exc)


class BasicLink(LinkCommon, CtxObj):
    """
    Simple direct link to a MQTT server.
    """

    def __init__(self, cfg, name: str | None, data: dict, **kw):
        super().__init__(cfg, name=name, **kw)
        self.data = data

    @asynccontextmanager
    async def _ctx(self):
        link = self.data["link"]
        if isinstance(link, dict):
            link = (link,)

        err = None
        yielded = False
        for remote in link:
            try:
                async with self._connect_one(remote, self.data) as self._sender:
                    yielded = True
                    yield self._sender
                    return
            except Exception as exc:
                if yielded:
                    raise
                self.logger.warning(
                    "Link failed: %r",
                    remote,
                    exc_info=None if isinstance(exc, EnvironmentError) else exc,
                )
                if err is None:
                    err = exc
            else:
                if not yielded and err is None:
                    err = EOFError

        if not yielded:
            if err is None:
                raise ValueError(f"No links in {self.data!r}")
            raise err


@define(eq=False)
class Watcher(CtxObj):
    """
    Helper class for monitoring (and coalescint the data of) either-or-both
    * a MQTT subscription to a subtree of our MoaT-Link hierarchy
    * a MoaT-Link request to enumerate a subtree
    """

    link: Link = field()
    path: Path = field()
    meta: bool = field()
    subtree: bool = field()
    state: bool | None | NotGiven = field()
    age: float | None = field()
    mark: bool = field()
    node_cls: type = field()
    min_length: type = field()
    max_length: type = field()

    _qr = field(init=False, repr=False)
    _tg = field(init=False, repr=False)
    _node = field(init=False, default=None)

    _current_done: anyio.Event | None = field(init=False, default=None)

    def __attrs_post_init__(self):
        if self.mark and self.state is NotGiven:
            raise ValueError("MQTT doesn't send a mark. Sorry.")
        if self.age is not None:
            self.age = time.time() - self.age

    def _chk(self, p, m):
        if self.age and m.timestamp < self.age:
            return False
        if self.min_length is not None and len(p) < self.min_length:
            return False
        if self.max_length is not None and len(p) > self.max_length:
            return False
        return True

    async def _current(self, qw, *, task_status):
        "get the current state from the server"
        if self.subtree:
            pl = PathLongener(())
            args = [self.path, self.age, self.min_length, self.max_length]
            while args[-1] is None:
                args.pop()
            async with self.link.d.walk(*args) as mon:
                task_status.started()
                async for r in mon:
                    n, p, d, *m = r
                    p = pl.long(n, p)
                    m = MsgMeta.restore(m)
                    if self._chk(p, m):
                        await qw.send((p, d, m))
                if self.mark:
                    await qw.send(None)
        else:
            try:
                r = await self.link.d.get(self.path)
            except (KeyError, ValueError):
                task_status.started()
            else:
                task_status.started()
                p, d, m = Path(), r[0], MsgMeta.restore(r[1:])
                if self._chk(p, m):
                    await qw.send((p, d, m))
            if self.mark:
                await qw.send(None)

        await qw.aclose()
        self._current_done.set()

    async def _updates(self, qw, *, task_status):
        "get updates from MQTT"
        plen = len(self.path)  # ignores the root prefix
        async with self.link.monitor(
            Root.get() + self.path, subtree=self.subtree, retained=(self.state is NotGiven)
        ) as mon:
            task_status.started()
            async for msg in mon:
                p, d, m = Path.build(msg.topic[plen:]), msg.data, msg.meta
                if self._chk(p, m):
                    await qw.send((p, d, m))
        await qw.aclose()

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            self._node = self.node_cls()
            self._current_done = anyio.Event()
            self._tg = tg
            qw, self._qr = anyio.create_memory_object_stream(10)
            if self.state is not True:
                await tg.start(self._updates, qw.clone())
            if self.state is True or self.state is None:
                await tg.start(self._current, qw.clone())
            else:
                self._current_done.set()
            await qw.aclose()
            yield self
            tg.cancel_scope.cancel()
            self._node = None

    def __aiter__(self):
        return self

    async def get_node(self, background=True) -> Node:
        """
        Wait until fetching the static data is complete, then return the
        watched node.

        If @background is set (the default), start a background job that
        updates the node. In this case you cannot iterate the watcher any
        more.
        """
        if background:
            await self._tg.start(self._iter)
        await self._current_done.wait()
        return self._node

    async def _iter(self, *, task_status: anyio.abc.TaskStatus):
        task_status.started()
        qr, self._qr = self._qr, None

        while True:
            try:
                msg = await qr.receive()
            except anyio.EndOfStream:
                return
            p, d, m = msg
            self._node.set(p, d, m)

    async def __anext__(self):
        while True:
            try:
                msg = await self._qr.receive()
            except anyio.EndOfStream:
                raise StopAsyncIteration from None
            if msg is None:
                return None
            p, d, m = msg
            if self._node.set(p, d, m, force=self._current_done.is_set()):
                if self.meta:
                    return (p, d, m) if self.subtree else (d, m)
                else:
                    return (p, d) if self.subtree else d


class Walker:
    """
    A trimmed-down watcher that retrieves a possibly-partial subtree
    from our MoaT-Link server.

    This differs from `Watcher` by not tracking updates, nor keeping a node
    tree in memory.
    """

    def __init__(self, mon, meta=False):
        self.mon = mon
        self.pl = PathLongener()
        self.meta = meta

    def __aiter__(self):
        return self

    async def __anext__(self):
        msg = await anext(self.mon)
        n, p, d, *m = msg.args
        p = self.pl.long(n, p)
        if self.meta:
            m = MsgMeta.restore(m)
            return p, d, m
        else:
            return p, d



"""
MoaT gateway
"""

from __future__ import annotations

import anyio

from moat.util import NotGiven, Path, PathLongener
from moat.kv.client import Client, open_client
from moat.link.meta import MsgMeta

from . import Gate as _Gate

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.link.gate import GateNode

    from typing import Any


class Gate(_Gate):  # noqa: D101
    kv: Client

    async def run_(self, *, task_status=anyio.TASK_STATUS_IGNORED):
        "Main loop. Overridden to start a Moat-KV client"
        async with open_client("moat.link.gate.kv", **self.cfg["kv"]) as self.kv:
            await super().run_(task_status=task_status)

    async def get_dst(self, task_status=anyio.TASK_STATUS_IGNORED):  # noqa: D102
        pl = PathLongener()
        # This chops the `self.cf.dst` prefix off the resulting path
        async with self.kv.watch(self.cf.dst, fetch=True, long_path=False, nchain=2) as mon:
            task_status.started()
            async for msg in mon:
                if "path" not in msg:
                    if msg.get("state", "") == "uptodate":
                        self.dst_is_current()
                    continue
                path = pl.long(msg.depth, msg.path)
                await self.set_src(
                    path,
                    msg.get("value", NotGiven),
                    MsgMeta(origin=msg.chain.node, t=msg.chain.tick),
                )

    async def set_dst(self, path: Path, data: Any, meta: MsgMeta, node: GateNode):
        "Set KV data."

        meta  # noqa:B018

        # XXX ideally we should have the previous value's external chain
        # available here, just to be able to complain when there's a conflict
        if data is NotGiven:
            res = await self.kv.delete(self.cf.dst + path, nchain=1)
        else:
            res = await self.kv.set(self.cf.dst + path, value=data, nchain=1)

        node.ext_meta = res.chain

    def is_update(self, node: GateNode, data: Any, meta: MsgMeta):
        "Check for update"
        data  # noqa:B018
        # If the message is an echo of what we sent earlier, ignore it.
        try:
            if meta.origin == node.ext_meta.node and meta["t"] == node.ext_meta.tick:
                return False
        except (AttributeError, KeyError):
            pass
        return True

    def newer_dst(self, node):  # noqa: D102
        # If the internal message has a copy of the outside metadata, it
        # should be either unmodified or older. Test the data to be sure.
        # Otherwise compare the chains.
        if node.meta.origin == self.origin:
            return None

        if "gw" in node.meta:
            last_node = node.meta["gw"]
            if last_node.origin != node.ext_meta.origin:
                return True
            if last_node["t"] < node.ext_meta["t"]:
                return True
            if last_node["t"] > node.ext_meta["t"]:
                return False
            return None

        # same data = nothing to do
        # XXX may require inexact comparison for floats
        if node.data == node.ext_data:
            return None

        # Otherwise assume that the local data is newer because there's no
        # gateway data in it.
        return False



# command line interface  # noqa: D100
from __future__ import annotations

import anyio
from contextlib import nullcontext

import asyncclick as click

from moat.util import MsgReader, NotGiven, P, attr_args, yprint
from moat.link._data import data_get, node_attr
from moat.link.client import Link
from moat.link.meta import MsgMeta


@click.group(short_help="Manage data.", invoke_without_command=True)  # pylint: disable=undefined-variable
@click.option("-m", "--meta", is_flag=True, help="include metadata")
@click.argument("path", type=P, nargs=1)
@click.pass_context
async def cli(ctx, path, meta):
    """
    This subcommand accesses the data stored in the MoaT-Link server.
    """
    obj = ctx.obj
    cfg = obj.cfg["link"]
    obj.conn = await ctx.with_async_resource(Link(cfg))
    obj.meta = meta
    if ctx.invoked_subcommand is None:
        await data_get(obj.conn, path, meta=obj.meta, out=obj.stdout, recursive=False)
    else:
        obj.path = path


@cli.command()
@click.option("-r", "--recursive", is_flag=True, help="Read a complete subtree")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=None,
    help="Limit recursion depth. Default: whole tree",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=None,
    help="Starting depth. Default: whole tree",
)
@click.option("-e", "--empty", is_flag=True, help="Include empty nodes")
@click.option("-R", "--raw", is_flag=True, help="Print string values without quotes etc.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.pass_obj
async def get(obj, **k):
    """
    Read a MoaT-KV value.

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    await data_get(obj.conn, obj.path, meta=obj.meta, **k)


@cli.command("list")
@click.option(
    "-d",
    "--as-dict",
    default=None,
    help="Structure as dictionary. The argument is the key to use "
    "for values. Default: return as list",
)
@click.option(
    "-m",
    "--maxdepth",
    type=int,
    default=1,
    help="Limit recursion depth. Default: 1 (single layer).",
)
@click.option(
    "-M",
    "--mindepth",
    type=int,
    default=1,
    help="Starting depth. Default: 1 (single layer).",
)
@click.pass_obj
async def list_(obj, **k):
    """
    List MoaT-KV values.

    This is like "get" but with "--mindepth=1 --maxdepth=1 --recursive --empty"

    If you read a sub-tree recursively, be aware that the whole subtree
    will be read before anything is printed. Use the "watch --state" subcommand
    for incremental output.
    """

    k["recursive"] = True
    k["raw"] = True
    k["empty"] = True
    await data_get(obj.conn, obj.path, meta=obj.meta, **k)


@cli.command("set", short_help="Add or update an entry")
@attr_args
@click.option("-l", "--last", nargs=2, help="Previous change entry (node serial)")
@click.option("-n", "--new", is_flag=True, help="This is a new entry.")
@click.pass_obj
async def set_(obj, last, new, **kw):
    """
    Store a value at some MoaT-KV position.

    If you update a value you can use "--last" to ensure that no other
    change arrived between reading and writing the entry.

    TODO: When adding a new entry use "--new" to ensure that you don't
    accidentally overwrite something.

    MoaT-KV entries typically are mappings. Use a colon as the path if you
    want to replace the top level.
    """
    last, new  # noqa:B018

    res = await node_attr(obj, obj.path, **kw)

    if obj.meta:
        yprint(res, stream=obj.stdout)


class nstr:  # noqa: D101
    def __new__(cls, val):  # noqa: D102
        if val is NotGiven:
            return val
        return str(val)


@cli.command(short_help="Delete an entry / subtree")
@click.option(
    "-b",
    "--before",
    type=float,
    help="Don't delete entries created after this timestamp",
)
@click.option("-r", "--recursive", is_flag=True, help="Delete a complete subtree")
@click.pass_obj
async def delete(obj, before, recursive):
    """
    Delete an entry, or a subtree.

    You really should use "--before" flag to ensure that no other change
    arrived after you viewed the data in question.

    Non-recursively deleting an entry with children works and does *not*
    affect the child entries.

    The root entry cannot be deleted.
    """
    args = {}
    if recursive:
        args["rec"] = recursive
    if before:
        args["ts"] = before
    res = await obj.conn.d.delete(obj.path, **args)
    if obj.meta:
        res = dict(data=res[0], meta=MsgMeta.restore(res[1:]).repr())
    else:
        res = res[0]
    yprint(res, stream=obj.stdout)


@cli.command()
@click.option("-m", "--mode", type=str, help="Retrieval mode", default="s")
@click.option("-M", "--mark", is_flag=True, help="Retrieval mode")
@click.option("-o", "--only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-s", "--subtree", is_flag=True, help="Read the whole tree.")
@click.option("-p", "--path-only", is_flag=True, help="Value only, nothing fancy.")
@click.option("-D", "--add-date", is_flag=True, help="Add *_date entries")
@click.option("-i", "--ignore", multiple=True, type=P, help="Skip this (sub)tree")
@click.option("-n", "--min-length", type=int, help="Minimum path length")
@click.option("-N", "--max-length", type=int, help="Maximum path length")
@click.option("-a", "--max-age", type=int, help="Skip entries older than N seconds")
@click.option("-t", "--timeout", type=int, help="Stop reading after N seconds")
@click.pass_obj
async def monitor(
    obj,
    mode,
    only,
    path_only,
    add_date,  # noqa: ARG001
    ignore,
    mark,
    subtree,
    min_length,
    max_length,
    max_age,
    timeout,
):
    """Monitor a MoaT-Link subtree.

    The mode can be:
    * c/current   read current data from the server
    * u/update    read updates from MQTT
    * s/stream    current plus updates
    * m/mqtt      subscribe to MQTT stream, including retained data
    """

    match mode:
        case "c" | "current":
            state = True
        case "u" | "update":
            state = False
        case "s" | "stream":
            state = None
        case "m" | "mqtt":
            state = NotGiven
        case _:
            raise click.UsageError("Mode must be current|update|stream|mqtt")
    if mark and state is not None:
        raise click.UsageError("You can only add a mark in Stream mode")

    def pm(p):
        for ip in ignore:
            if len(p) >= len(ip) and p[: len(ip)] == ip:
                return True
        return False

    with anyio.move_on_after(timeout) if timeout else nullcontext():
        async with obj.conn.d_watch(
            obj.path,
            state=state,
            mark=mark,
            meta=True,
            subtree=subtree,
            max_length=max_length,
            min_length=min_length,
            age=max_age,
        ) as mon:
            async for pdm in mon:
                if pdm is None:
                    res = "*** Snapshot data ends ***"
                else:
                    p, d, m = pdm
                    if pm(p):
                        continue

                    m = MsgMeta.restore(m)
                    if only:
                        res = d
                    elif path_only:
                        res = p
                    elif obj.meta:
                        res = [p, d, m]
                    else:
                        res = [p, d]
                yprint(res, stream=obj.stdout)
                print("---", file=obj.stdout)
                obj.stdout.flush()


@cli.command()
@click.option("-i", "--infile", type=click.Path(), help="File to read.")
@click.option("-C", "--codec", type=str, default="yaml", help="Codec to use (default: yaml).")
@click.pass_obj
async def update(obj, infile, codec):
    """Write a list of updates to a MoaT-Link subtree"""
    async with MsgReader(path="/dev/stdin" if infile == "-" else infile, codec=codec) as reader:
        async for msg in reader:
            if isinstance(msg, dict):
                p = msg["path"]
                v = msg["value"]
            else:
                p, v, *_m = msg
            await obj.conn.d_set(obj.path + p, v)



"""
Common code for backends.

Currently the only backend supported is MQTT, but who knows.
"""

from __future__ import annotations

import anyio
import logging
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager

from attrs import define, field

from moat.util import CtxObj, NotGiven, Path, Root, attrdict, get_codec

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mqttproto import QoS

    from moat.lib.codec import Codec
    from moat.link.meta import MsgMeta

    from collections.abc import AsyncIterator
    from typing import Any, ClassVar, Literal, Self


__all__ = ["Backend", "Message", "RawMessage", "get_backend", "get_codec"]


@define
class Message[TData]:
    """
    An incoming message.
    """

    topic: Path = field()
    data: TData = field()
    meta: MsgMeta = field()
    prop: dict[str, Any] = field(repr=False)
    retain: bool | None = field(default=False)

    raw: ClassVar[bool] = False

    def __class_getitem__(cls, TData):
        return cls  # for now


@define
class RawMessage(Message):
    "A message that couldn't be decoded / shouldn't be encoded"

    exc: Exception = field(default=None)
    raw: ClassVar[bool] = True


class Backend(CtxObj, metaclass=ABCMeta):
    "Base class for messaging backends"

    _tg: anyio.abc.TaskGroup
    _njobs: int = 0
    _ended: anyio.Event | None = None

    def __init__(self, cfg: attrdict, name: str, id: str):
        self.cfg = cfg
        self.name = name
        self.id = id
        self.logger = logging.getLogger(f"moat.link.backend.{name}")

    @abstractmethod
    @asynccontextmanager
    async def connect(self, *a, **k) -> AsyncIterator[Self]:
        """
        This async context manager returns a connection to the backend.
        """

    @asynccontextmanager
    async def _ctx(self) -> AsyncIterator[Self]:
        async with (
            anyio.create_task_group() as self._tg,
            self.connect(),
        ):
            yield self

    @abstractmethod
    @asynccontextmanager
    async def monitor(
        self,
        topic: Path,
        qos: QoS | None = None,
        codec: Codec | None | Literal[NotGiven] = NotGiven,
        raw: bool | None = False,
        echo: bool = False,
        no_local: bool = False,
        subtree: bool = False,
    ) -> AsyncIterator[AsyncIterator[Message]]:
        """
        Return an async iterator that listens to this topic.

        Set @raw to ``True`` to always get non-decoded messages.
        """

    @abstractmethod
    async def send(
        self,
        topic: Path,
        data: Any,
        codec: Codec | None | Literal[NotGiven] = NotGiven,
        **kw: Any,
    ) -> None:
        """
        Send this payload to this topic.
        """

    async def send_error(
        self,
        subpath: Path,
        msg: str | None = None,
        data: Any = None,
        exc: Exception | None = None,
        **kw: Any,
    ):
        """
        Send a somewhat-free-form error message.
        """
        if subpath.prefix is None:
            subpath = Root.get() / "error" + subpath
        if msg:
            kw["msg"] = msg
        if data is not NotGiven:
            kw["data"] = data
        if exc is not None:
            kw["exc"] = repr(exc)
        # TODO include the backtrace?

        try:
            await self.send(subpath, kw, codec="std-cbor")
        except Exception:
            self.logger.exception("Failure to log error at %s: %r", subpath, data, exc_info=exc)
        else:
            self.logger.debug("Error at %s: %r", subpath, data, exc_info=exc)


def get_backend(cfg: dict, **kw) -> Backend:
    """
    Fetch the backend named in the config and initialize it.
    """
    from importlib import import_module  # noqa: PLC0415

    cfg = cfg["backend"]
    name = cfg["driver"]
    if "." not in name:
        name = "moat.link.backend." + name
    return import_module(name).Backend(cfg, **kw)



"""
Data access
"""

from __future__ import annotations

import anyio
import os
import sys
import time

from moat.util import NotGiven, Path, PathLongener, attrdict, process_args, yprint
from moat.util.times import ts2iso

from .meta import MsgMeta

from collections.abc import Mapping
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.codec import Codec
    from moat.link.client import Link


def add_dates(d):
    """
    Given a dict with int/float entries that might conceivably be dates,
    add ``_*`` with a textual representation.
    """

    t = time.time()
    start = t - 366 * 24 * 3600
    stop = t + 366 * 24 * 3600

    def _add(d):
        if isinstance(d, (list, tuple)):
            for dd in d:
                _add(dd)
            return
        if not isinstance(d, Mapping):
            return
        for k, v in list(d.items()):
            if isinstance(k, str) and k.startswith("_"):
                continue
            if not isinstance(v, (int, float)):
                _add(v)
                continue
            if start <= v <= stop:
                d[f"_{k}"] = ts2iso(v)

    _add(d)


async def data_get(
    conn: Link,
    path: Path,
    *,
    meta: bool = False,
    recursive: bool = True,
    as_dict: str | None = "_",
    maxdepth: int = -1,
    mindepth: int = 0,
    empty: bool = False,
    raw: bool = False,
    path_mangle=None,
    item_mangle=None,
    add_date=False,
    out=None,
):
    """Generic code to dump a subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    if path_mangle is None:

        def path_mangle(x):
            return x

    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if out is None:
        out = sys.stdout
    elif out is False:
        out = []

    if recursive:
        kw = {}
        a = [None, None, None]

        if maxdepth is not None and maxdepth >= 0:
            a[2] = maxdepth
        if mindepth:
            a[1] = mindepth
        if empty:
            kw["empty"] = True
        while a and a[-1] is None:
            a.pop()
        y = {}
        pl = PathLongener()
        async with conn.d.walk(path, *a, **kw).stream_in() as res:
            async for r in res:
                r = await item_mangle(r)  # noqa:PLW2901
                if r is None:
                    continue
                n, p, d, *m = r
                path = pl.long(n, p)
                path = path_mangle(path)
                if path is None:
                    continue

                if add_date:
                    add_dates(d)

                if meta:
                    m = MsgMeta._moat__restore(m, NotGiven)  # noqa: SLF001

                if as_dict is not None:
                    if meta:
                        d = dict(data=d, meta=m.repr())
                    yy = y
                    for p in path:
                        yy = yy.setdefault(p, {})
                    try:
                        yy[as_dict] = d
                    except AttributeError:
                        if empty:
                            yy[as_dict] = None
                else:
                    if raw:
                        y = path
                    elif meta:
                        y = [path, d, m.repr()]
                    else:
                        y = [path, d]
                    if isinstance(out, list):
                        out.append(y)
                    else:
                        yprint(y, stream=out)
                        out.write("---\n")

            if as_dict is not None:
                if maxdepth:

                    def simplex(d):
                        for k, v in d.items():
                            if isinstance(v, dict):
                                d[k] = simplex(d[k])
                        if as_dict in d and d[as_dict] is None:
                            if len(d) == 1:
                                return None
                            else:
                                del d[as_dict]
                        return d

                    y = simplex(y)

            if isinstance(out, list):
                return y
            if as_dict is not None:
                yprint(y, stream=out)

            return out  # end "if recursive"

    d, *m = await conn.d.get(path)
    if add_date:
        add_dates(d)
    if meta:
        m = MsgMeta.restore(m)
        d = dict(data=d, meta=m.repr())

    if out is False:
        return d
    if not raw:
        yprint(d, stream=out)
    elif isinstance(d, bytes):
        os.write(out.fileno(), res)
    else:
        out.write(str(d))
    pass  # end get


async def backend_get(
    conn: Link,
    path: Path,
    *,
    meta: bool = False,
    recursive: bool = True,
    empty: bool = False,
    as_dict: str | None = "_",
    raw: bool = False,
    path_mangle=None,
    item_mangle=None,
    add_date=False,
    codec: Codec | None = None,
    out=None,
    timeout: float = 0.5,
):
    """Generic code to dump a backend subtree.

    `path_mangle` accepts a path and the as_dict parameter. It should
    return the new path. This is used for e.g. prefixing the path with a
    device name. Returning ``None`` causes the entry to be skipped.
    """
    # This is a copy of `data_get` that accesses the backend directly.

    if path_mangle is None:

        def path_mangle(x):
            return x

    if item_mangle is None:

        async def item_mangle(x):  # pylint: disable=function-redefined
            return x

    if out is None:
        out = sys.stdout
    elif out is False:
        out = []

    if recursive:
        kw = {"codec": codec}

        y = {}
        async with conn.monitor(path, subtree=True, **kw) as mon:
            while True:
                r = None
                with anyio.move_on_after(timeout):
                    r = await anext(mon)
                if r is None:
                    break

                r = await item_mangle(r)
                if r is None:
                    continue
                p, d, m = r.topic, r.data, r.meta
                p = p[len(path) :]
                p = path_mangle(p)
                if p is None:
                    continue

                if add_date:
                    add_dates(d)
                if meta:
                    m = MsgMeta._moat__restore(m, NotGiven)  # noqa: SLF001
                    d = dict(data=d, meta=m.repr())

                if as_dict is not None:
                    yy = y
                    for pp in p:
                        yy = yy.setdefault(pp, {})
                    try:
                        yy[as_dict] = d
                    except AttributeError:
                        if empty:
                            yy[as_dict] = None
                else:
                    if raw:
                        y = p
                    else:
                        y = {}
                        try:
                            y[p] = d
                        except AttributeError:
                            if empty:
                                y[p] = None
                            else:
                                continue
                    if isinstance(out, list):
                        out.append(y)
                    else:
                        yprint([y], stream=out)

            if isinstance(out, list):
                return y
            yprint(y, stream=out)

            return out  # end "if recursive"

    async with conn.monitor(path, **kw) as mon:
        r = None
        with anyio.move_on_after(timeout):
            r = await anext(mon)
        if r is None:
            raise KeyError(path)

    d, m = r.data, r.meta
    if add_date:
        add_dates(d)
    if meta:
        m = MsgMeta.restore(m)
        d = dict(data=d, meta=m.repr())

    if out is False:
        return d
    if not raw:
        yprint(d, stream=out)
    elif isinstance(d, bytes):
        os.write(out.fileno(), d)
    else:
        out.write(str(d))
    pass  # end get


def res_get(res, attr: Path, **kw):  # pylint: disable=redefined-outer-name
    """
    Get a node's value and access the dict items beneath it.

    The node value must be an attrdict.
    """
    val = res.get("value", None)
    if val is None:
        return None
    return val._get(attr, **kw)  # noqa: SLF001


def res_update(res, attr: Path, value=None, **kw):  # pylint: disable=redefined-outer-name
    """
    Set a node's sub-item's value, possibly merging dicts.
    Entries set to 'NotGiven' are deleted.

    The node value must be an attrdict.

    Returns the new value.
    """
    val = res.get("value", attrdict())
    return val._update(attr, value=value, **kw)  # noqa: SLF001


async def node_attr(obj, path, val=NotGiven, meta=NotGiven, **kw):
    """
    Sub-attr setter.

    Args:
        obj: command object
        path: address of the node to change
        res: old node, if it has been read already
        **kw: the results of `attr_args`

    Returns the result of setting the attribute.
    """
    if val is NotGiven:
        try:
            val, *m = await obj.conn.d.get(path)
        except KeyError:
            pass
        else:
            meta = MsgMeta.restore(m)
    val = process_args(val, **kw)
    t = {} if meta is NotGiven else {"t": meta.timestamp}
    if val is NotGiven:
        res = await obj.conn.d.delete(path, **t)
    else:
        res = await obj.conn.d.set(path, val, **t)
    res = res[0], MsgMeta.restore(res[1:])
    return res



"""
This module contains all MoaT-Link exceptions.

Currently this is (mostly) a re-import of moat.kv.exceptions.
"""

from __future__ import annotations


class MoaTLinkError(RuntimeError):
    """Superclass of all MoaT-KV errors.

    Abstract class.
    """

    pass


class ClientError(MoaTLinkError):  # noqa: D101
    pass


class ClientCancelledError(ClientError):  # noqa: D101
    pass


class AuthError(MoaTLinkError):  # noqa: D101
    pass


class OutOfDateError(MoaTLinkError):  # noqa: D101
    pass


class ServiceSupplanted(MoaTLinkError):  # noqa: D101
    pass


class ServiceNotFound(MoaTLinkError):  # noqa: D101
    pass


class ServiceNotStarted(UserWarning):  # noqa: D101
    pass



"""
Common parts
"""

from __future__ import annotations

import anyio

from moat.lib.cmd.base import MsgHandler

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.cmd import Msg

    from collections.abc import Awaitable


class NotAuthorized(RuntimeError):  # noqa: D101
    pass


class CmdCommon(MsgHandler):
    """
    Commands every handler should know.

    This includes the 'i' subcommand.
    """

    doc_i = dict(_d="Internal commands")

    def sub_i(self, msg: Msg, rcmd: list) -> Awaitable[None]:
        "Local subcommand redirect for 'i'"
        return self.handle(msg, rcmd, "i")

    doc_i_ping = dict(_d="Ping, echo", _r="Any: sends all args and keys back")
    doc_i_乒 = doc_i_ping  # noqa:PLC2401

    async def stream_i_ping(self, msg: Msg) -> bool | None:
        """
        This handler replies with "pong" and its arguments, for basic
        round-trip tests.

        乒 ⇒ 乓

        Yes, this name is silly.
        """
        await msg.result("乓", *msg.args, **msg.kw)

    stream_i_乒 = stream_i_ping  # noqa:PLC2401

    doc_i_count = dict(
        _d="count", _o="int:numbers", s="int:start", d="int:delta", e="int:end", t="float:interval"
    )

    async def stream_i_count(self, msg: Msg) -> bool | None:
        """
        This handler replies with a stream of numbers.

        The result is the first value that crossed the line.
        """
        s = msg.get("s", 0)
        e = msg.get("e", None)
        d = msg.get("d", None)
        t = msg.get("t", 1)

        async with msg.stream_out():
            if e is not None and e < s:
                if d is None:
                    d = -1
                while e < s:
                    await msg.send(s)
                    s += d
                    await anyio.sleep(t)
            else:
                if d is None:
                    d = 1
                while e is None or e > s:
                    await msg.send(s)
                    s += d
                    await anyio.sleep(t)
            await msg.result(s)

    doc_i_sink = dict(_d="count up", _i="Any:whatever", n="int:accept this many")

    async def stream_i_sink(self, msg: Msg) -> bool | None:
        """
        This handler swallows an input stream.

        @n is the number of messages that shall be accepted.

        The result is the numebr of messages that arrived.
        """
        n = msg.get("n", -1)
        c = 0

        async with msg.stream_in() as st:
            while True:
                if c == n:
                    break
                _m = await anext(st)

            await msg.result(c)

    doc_i_echo = dict(
        _d="count up", _i="Any:whatever", _o="Any:whatever", n="int:accept this many"
    )

    async def stream_i_echo(self, msg: Msg) -> bool | None:
        """
        This handler echoes an input stream.

        @n is the number of messages that shall be accepted.

        The result is the number of messages that arrived.
        """
        n = msg.get("n", -1)
        c = 0

        async with msg.stream() as st:
            while True:
                if c == n:
                    break
                m = await anext(st)
                await msg.send(m)

            await msg.result(c)
